import os

import numpy as np

class PowerData(object):
    """
    Dummy object for calculating a likelihood
    """
    def __init__(self, data_params):
        """
        Constructor of the power spectrum data
        """
        # load the power spectrum for all tracer combinations
        power = {}
        for key in data_params['tracer_combos'].keys():
            power[key] = np.load(data_params['tracer_combos'][key]['path2power'])['xi']
        self.power = power

        # load the covariance matrix for all tracer combinations
        icov = {}
        for key in data_params['tracer_combos'].keys():
            cov = np.load(data_params['tracer_combos'][key]['path2cov'])['xicov']
            icov[key] = np.linalg.inv(cov)
        self.icov = icov


    def compute_likelihood(self, theory):
        """
        Computes the likelihood using information from the context
        """
        # Calculate a likelihood up to normalization
        lnprob = 0.
        for key in self.power.keys():
            delta = (self.power[key] - theory[key]).flatten()
            lnprob += np.einsum('i,ij,j',delta, self.icov[key], delta)
        lnprob *= -0.5

        # Return the likelihood
        print(" <><> Likelihood evaluated, lnprob = ",lnprob)
        return lnprob



#!/usr/bin/env python2
from __future__ import print_function
import argparse, base64, hashlib, json, os, SimpleHTTPServer, socket, SocketServer, struct as S, sys
from threading import current_thread, Lock, Thread
import pkg_resources

try:
    import .kvsclient
    from .kvscommon import recvall
except:
    import kvsclient
    from kvscommon import recvall

# The web monitor implements two conventions wrt keys:
#
# 1) A key starting with '.' is not displayed.
#
# 2) A wait for a key of the form "key?fmt" creates an input
# box for the key. fmt will be used to format the input string
# to produce a "put" value for the key.
#
# It implements three conventions wrt to values:
#
# 1) An ASCII String encoded value that starts with '<HTML>' (case
# insensitive) will be sent directly to the web client after removing
# the '<HTML>' pefix. I.e., this combination permits the value to be
# render as HTML.
#
# 2) An ASCII String encoded value that does not start with '<HTML>'
# is sent HTML escaped and encoded (distinct from the KVS encoding
# scheme). This combination prevents the value from being interpreted
# as HTML.
#
# 3) In all other cases, python's 'repr' of the (KVS encoding, value)
# tuple is sent to the web client HTML escaped and encoded.
#

def jsonDefault(o):
    if isinstance(o, bytearray):
        return o.decode('latin-1')
    return repr(o)

def dump2json(kvsc):
    r = json.dumps(kvsc.dump(), ensure_ascii = False, check_circular = False, encoding = 'latin-1', default = jsonDefault)
    if isinstance(r, unicode): r = r.encode('latin-1')
    return r

class KVSWaitThread(Thread):
    def __init__(self, kvsaddr, wslist, mk, spec, frontend, name='KVSClientThread'):
        Thread.__init__(self, name=name)
        self.daemon = True
        self.mk = mk
        self.wslist = wslist
        self.kvsc = kvsclient.KVSClient(kvsaddr)
        self.kvsc.monkey(mk, spec)
        self.frontend = frontend
        self.start()

    def run(self):
        try: 
            while 1:
                r = self.kvsc.get(self.mk, False)
                j = dump2json(self.kvsc)
                self.wslist.broadcast(j)
        finally:
            self.wslist.broadcast('bye', True)
            self.kvsc.close()
            self.frontend.stop()

class WebSocketServer(object):
    def __init__(self, kvsaddr, wslist, ws):
        self.client = ws
        self.lock = Lock()
        self.active = True

        kvsc = kvsclient.KVSClient(kvsaddr)
        wslist.add(self)

        try:
            while self.active:
                op = self.recv()
                if not op: pass
                elif op == 'bye': break
                elif op == 'dump':
                    j = dump2json(kvsc)
                    self.send(j)
                elif op.startswith('put\x00'):
                    d, key, v = op.split('\x00')
                    if '?' in key:
                        k, fmt = key.split('?')
                        if fmt: v = fmt%v
                    kvsc.put(key, v, False)
                else: raise Exception('Unknown op from websocket: "%s".'%repr(op))
        finally:
            wslist.remove(self)
            kvsc.close()
            self.close()

    def send(self, p, end = False):
        try:
            with self.lock:
                op = 8 if end else 1
                lp = len(p)
                if lp < 126:
                    self.client.sendall(S.pack('BB', 0x80 | op, lp))
                elif lp < 2**16:
                    self.client.sendall(S.pack('!BBH', 0x80 | op, 126, lp))
                else:
                    self.client.sendall(S.pack('!BBQ', 0x80 | op, 127, lp))
                self.client.sendall(p)
        except socket.error as msg:
            self.active = False

    def recv(self):
        r = self.client.recv(1)
        if not r: return 'bye'
        b = S.unpack('B', r)[0]
        op = b & 0xF
        if op == 0x8:
            assert b & 0x80
            return 'bye'
        elif op == 0x9:
            print('Got a ping', file=sys.stderr)
            #TODO: Do something here
            return ''
        elif op == 0XA:
            print('Got a pong', file=sys.stderr)
            #TODO: Do something here?
            return ''
        elif op == 0x1:
            assert b & 0x80 # For this test, this frame must be marked FIN
            b = S.unpack('B', recvall(self.client, 1))[0]
            assert b & 0x80 # Masking must be on for frames from clients.
            plen = b & 0x7F
            assert plen < 126 # Limit on payload for this test.
            mb = S.unpack('4B', recvall(self.client, 4))
            p = recvall(self.client, plen)
            decode = ''
            for i in xrange(plen): decode += chr(ord(p[i]) ^ mb[i%4])
            return decode
        else:
            print('ws recv header byte: %02x'%b, file=sys.stderr)
            return ''

    def close(self):
        self.client.close()

class WebSocketList(object):
    def __init__(self):
        self.lock = Lock()
        self.wss = set() # of WebWaitThread
    
    def add(self, ws):
        with self.lock:
            self.wss.add(ws)

    def remove(self, ws):
        with self.lock:
            self.wss.remove(ws)

    def __iter__(self):
        with self.lock:
            return iter(self.wss.copy())

    def broadcast(self, p, end = False):
        for ws in self:
            # TODO: add timeout for sending to avoid blocking, or lift to separate thread
            ws.send(p, end)

class FrontEnd(SimpleHTTPServer.SimpleHTTPRequestHandler):
    html = pkg_resources.resource_filename('kvsstcp', 'wskvspage.html')

    wsmagic = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'

    handshake = '\
HTTP/1.1 101 Switching Protocols\r\n\
Upgrade: websocket\r\n\
Connection: Upgrade\r\n\
Sec-WebSocket-Accept: %s\r\n\r\n\
'

    def send_head(self):
        if self.headers.get('upgrade', None) == 'websocket' and self.headers.get('sec-websocket-protocol', None) == 'kvs' and self.headers.get('origin') == self.server.base_url:
            k = self.headers.get('sec-websocket-key')
            reply = FrontEnd.handshake%(base64.b64encode(hashlib.sha1(k + FrontEnd.wsmagic).digest()))
            self.request.sendall(reply)
            WebSocketServer(self.server.kvsserver, self.server.wslist, self.request)
            return None

        elif self.path == '/kvsviewer':
            f = open(FrontEnd.html, 'r')
            self.send_response(200)
            self.send_header('Content-type', 'text/html; charset=utf-8')
            fs = os.fstat(f.fileno())
            self.send_header("content-length", str(fs.st_size))
            self.send_header("last-modified", self.date_time_string(fs.st_mtime))
            self.end_headers()
            return f
            
        else:
            self.send_error(404, 'Not recognized')
            return None

class ThreadingTCPServer(SocketServer.ThreadingTCPServer):
    daemon_threads = True

class FrontEndThread(Thread):
    def __init__(self, kvsserver, addr, urlfile, wslist, name='FrontEndThread'):
        Thread.__init__(self, name=name)
        self.daemon = True
        self.httpd = ThreadingTCPServer(addr, FrontEnd)
        if addr[1]: self.httpd.allow_reuse_address = True
        self.httpd.base_url = 'http://%s:%s'%(self.httpd.server_address)

        myurl = self.httpd.base_url + '/kvsviewer'
        print('front end at: '+myurl, file=sys.stderr)
        if urlfile:
            urlfile.write('%s\n'%(myurl))
            urlfile.close()

        self.httpd.kvsserver = kvsserver
        self.httpd.wslist = wslist
        self.start()

    def run(self):
        self.httpd.serve_forever()
        self.httpd.server_close()

    def stop(self):
        self.httpd.shutdown()

def main(kvsserver, urlfile=None, monitorkey='.webmonitor', monitorspec=':w', addr=(socket.gethostname(), 0), **args):
    wslist = WebSocketList()
    fe = FrontEndThread(kvsserver, addr, urlfile, wslist)
    kvs = KVSWaitThread(kvsserver, wslist, monitorkey, monitorspec, fe)
    return (fe, kvs)

if '__main__' == __name__:
    argp = argparse.ArgumentParser(description='Start a web monitor for a key-value storage server.')
    argp.add_argument('-m', '--monitorkey', default='.webmonitor', help='Key to use for the monitor.')
    argp.add_argument('-s', '--monitorspec', default=':w', help='What to monitor: comma separted list of "[key]:[gpvw]" specifications.')
    argp.add_argument('-u', '--urlfile', default=None, type=argparse.FileType('w'), help='Write url to this file.')
    argp.add_argument('-b', '--bind', default=socket.gethostname(), type=str, help='HTTP IP to listen on (hostname).')
    argp.add_argument('-p', '--port', default=0, type=int, help='HTTP port to listen on (random).')
    kvsclient.addKVSServerArgument(argp)
    args = argp.parse_args()

    args.addr = (args.bind, args.port)
    (fe, kvs) = main(**args.__dict__)
    try:
        while fe.isAlive():
            fe.join(60)
    finally:
        fe.stop()
    fe.join()



#!/usr/bin/python2
from __future__ import print_function

import os, time, sys
from kvsclient import KVSClient

kvs = KVSClient(os.getenv('KVSSTCP_HOST'), os.getenv('KVSSTCP_PORT'))
rank, pop = int(os.getenv('SLURM_PROCID')), int(os.getenv('SLURM_NTASKS'))

myTag = 'data%10d'%rank
pTag = 'data%10d'%((rank+1)%pop)

Loops = 100
for payloadSize in [1<<x for x in range(21)]:
    p = 'a'*(payloadSize)
    t0 = time.time()
    for l in range(Loops):
        kvs.put(myTag, p,  False)
        p = kvs.get(pTag, False)
    t1 = time.time()
    delta = t1 - t0
    print("%20d%20.2e%20.2e%20.2e"%(payloadSize, delta/Loops, pop*Loops*2/delta, pop*payloadSize*Loops*2/delta))

print(kvs.dump())





#!/usr/bin/env python
from distutils.core import setup
setup(name='disbatch', version='0',
    description="Distributed processing of a batch of tasks",
    long_description=open("Readme.md").read(),
    requires=['kvsstcp'],
    scripts=['disBatch.py'])



#!/usr/bin/env python3
'''
This is a script for loading simulation data and generating subsamples.

Usage
-----
$ python -m abacusnbody.hod.AbacusHOD.prepare_sim --path2config /path/to/config.yaml
'''

import os
from pathlib import Path
import yaml

import numpy as np
import random
import time
from astropy.table import Table
import h5py
from scipy.ndimage import gaussian_filter
from scipy.interpolate import NearestNDInterpolator
from itertools import repeat
import argparse

from abacusnbody.data.compaso_halo_catalog import CompaSOHaloCatalog

import multiprocessing
from multiprocessing import Pool

from sklearn.neighbors import KDTree

DEFAULTS = {}
DEFAULTS['path2config'] = 'config/abacus_hod.yaml'


# https://arxiv.org/pdf/2001.06018.pdf Figure 13 shows redshift evolution of LRG HOD 
# the subsampling curve for halos
def subsample_halos(m, MT):
    x = np.log10(m)
    if MT:
        downfactors = np.zeros(len(x))
        mask = x < 12.07
        downfactors[mask] = 0.2/(1.0 + 10*np.exp(-(x[mask] - 11.2)*25))
        downfactors[~mask] = 1.0/(1.0 + 0.1*np.exp(-(x[~mask] - 12.6)*7))
        return downfactors
    else:
        return 1.0/(1.0 + 0.1*np.exp(-(x - 12.6)*7)) # LRG only

def subsample_particles(m, MT):
    x = np.log10(m)
    # return 4/(200.0 + np.exp(-(x - 13.7)*8)) # LRG only
    if MT:
        return 0.03 # 4/(200.0 + np.exp(-(x - 13.2)*6)) # MT
    else:
        return 4/(200.0 + np.exp(-(x - 13.7)*8)) # LRG only

# # these two functions are for grid based density calculation. We found the grid based definiton is disfavored by data
# def get_smo_density_oneslab(i, simdir, simname, z_mock, N_dim, cleaning):
#     slabname = simdir+simname+'/halos/z'+str(z_mock).ljust(5, '0')\
#     +'/halo_info/halo_info_'+str(i).zfill(3)+'.asdf'

#     cat = CompaSOHaloCatalog(
#         slabname, fields = ['N', 'x_L2com'], cleaned = cleaning)
#     Lbox = cat.header['BoxSizeHMpc']
#     halos = cat.halos

#     if cleaning:
#         halos = halos[halos['N'] > 0]

#     # get a 3d histogram with number of objects in each cell                                                                       
#     D, edges = np.histogramdd(halos['x_L2com'], weights = halos['N'],
#         bins = N_dim, range = [[-Lbox/2, Lbox/2],[-Lbox/2, Lbox/2],[-Lbox/2, Lbox/2]])   
#     return D


# def get_smo_density(smo_scale, numslabs, simdir, simname, z_mock, N_dim, cleaning):   
#     Dtot = 0
#     for i in range(numslabs):
#         Dtot += get_smo_density_oneslab(i, simdir, simname, z_mock, N_dim, cleaning)   

#     # gaussian smoothing 
#     Dtot = gaussian_filter(Dtot, sigma = smo_scale, mode = "wrap")

#     # average number of particles per cell                                                                                         
#     D_avg = np.sum(Dtot)/N_dim**3                                                                                                                                                                                                                              
#     return Dtot / D_avg - 1
    
def prepare_slab(i, savedir, simdir, simname, z_mock, tracer_flags, MT, want_ranks, want_AB, cleaning, newseed):
    outfilename_halos = savedir+'/halos_xcom_'+str(i)+'_seed'+str(newseed)+'_abacushod_oldfenv'
    outfilename_particles = savedir+'/particles_xcom_'+str(i)+'_seed'+str(newseed)+'_abacushod_oldfenv'
    print("processing slab ", i)
    if MT:
        outfilename_halos += '_MT'
        outfilename_particles += '_MT'
    if want_ranks:
        outfilename_particles += '_withranks'
    outfilename_particles += '_new.h5'
    outfilename_halos += '_new.h5'

    np.random.seed(newseed + i)
    # if file already exists, just skip
    # if os.path.exists(outfilename_halos) \
    # and os.path.exists(outfilename_particles):
    #     return 0

    # load the halo catalog slab
    print("loading halo catalog ")
    slabname = simdir+simname+'/halos/z'+str(z_mock).ljust(5, '0')\
    +'/halo_info/halo_info_'+str(i).zfill(3)+'.asdf'

    cat = CompaSOHaloCatalog(slabname, subsamples=dict(A=True, rv=True), fields = ['N', 
        'x_L2com', 'v_L2com', 'r90_L2com', 'r25_L2com', 'r98_L2com', 'npstartA', 'npoutA', 'id', 'sigmav3d_L2com'], 
        cleaned = cleaning)
    halos = cat.halos
    if cleaning:
        halos = halos[halos['N'] > 0]

    parts = cat.subsamples
    header = cat.header
    Lbox = cat.header['BoxSizeHMpc']
    Mpart = header['ParticleMassHMsun'] # msun / h 
    H0 = header['H0']
    h = H0/100.0

    # # form a halo table of the columns i care about 
    # creating a mask of which halos to keep, which halos to drop
    p_halos = subsample_halos(halos['N']*Mpart, MT)
    mask_halos = np.random.random(len(halos)) < p_halos
    print("total number of halos, ", len(halos), "keeping ", np.sum(mask_halos))

    halos['mask_subsample'] = mask_halos
    halos['multi_halos'] = 1.0 / p_halos

    # only generate fenv ranks and c ranks if the user wants to enable secondary biases
    if want_AB:
        nbins = 100
        mbins = np.logspace(np.log10(3e10), 15.5, nbins + 1)

        # # grid based environment calculation
        # dens_grid = np.array(h5py.File(savedir+"/density_field.h5", 'r')['dens'])
        # ixs = np.floor((np.array(halos['x_L2com']) + Lbox/2) / (Lbox/N_dim)).astype(np.int) % N_dim
        # halos_overdens = dens_grid[ixs[:, 0], ixs[:, 1], ixs[:, 2]]
        # fenv_rank = np.zeros(len(halos))
        # for ibin in range(nbins):
        #     mmask = (halos['N']*Mpart > mbins[ibin]) & (halos['N']*Mpart < mbins[ibin + 1])
        #     if np.sum(mmask) > 0:
        #         if np.sum(mmask) == 1:
        #             fenv_rank[mmask] = 0
        #         else:
        #             new_fenv_rank = halos_overdens[mmask].argsort().argsort()
        #             fenv_rank[mmask] = new_fenv_rank / np.max(new_fenv_rank) - 0.5
        # halos['fenv_rank'] = fenv_rank

        allpos = halos['x_L2com']
        allmasses = halos['N']*Mpart
        allpos_tree = KDTree(allpos)

        allinds_inner = allpos_tree.query_radius(allpos, r = halos['r98_L2com'])
        allinds_outer = allpos_tree.query_radius(allpos, r = 5)
        print("computng m stacks")
        starttime = time.time()
        Menv = np.array([np.sum(allmasses[allinds_outer[ind]]) - np.sum(allmasses[allinds_inner[ind]]) \
            for ind in np.arange(len(halos))])

        fenv_rank = np.zeros(len(Menv))
        for ibin in range(nbins):
            mmask = (halos['N']*Mpart > mbins[ibin]) \
            & (halos['N']*Mpart < mbins[ibin + 1])
            if np.sum(mmask) > 0:
                if np.sum(mmask) == 1:
                    fenv_rank[mmask] = 0
                else:
                    new_fenv_rank = Menv[mmask].argsort().argsort()
                    fenv_rank[mmask] = new_fenv_rank / np.max(new_fenv_rank) - 0.5

        halos['fenv_rank'] = fenv_rank

        # compute delta concentration
        print("computing c rank")
        halos_c = halos['r90_L2com']/halos['r25_L2com']
        deltac_rank = np.zeros(len(halos))
        for ibin in range(nbins):
            mmask = (halos['N']*Mpart > mbins[ibin]) & (halos['N']*Mpart < mbins[ibin + 1])
            if np.sum(mmask) > 0:
                if np.sum(mmask) == 1:
                    deltac_rank[mmask] = 0
                else:
                    new_deltac = halos_c[mmask] - np.median(halos_c[mmask])
                    new_deltac_rank = new_deltac.argsort().argsort()
                    deltac_rank[mmask] = new_deltac_rank / np.max(new_deltac_rank) - 0.5
        halos['deltac_rank'] = deltac_rank

    else:
        halos['fenv_rank'] = np.zeros(len(halos))
        halos['deltac_rank'] = np.zeros(len(halos))

    # the new particle start, len, and multiplier
    halos_pstart = halos['npstartA']
    halos_pnum = halos['npoutA']
    halos_pstart_new = np.zeros(len(halos))
    halos_pnum_new = np.zeros(len(halos))

    # particle arrays for ranks and mask 
    mask_parts = np.zeros(len(parts))
    len_old = len(parts)
    ranks_parts = np.full(len_old, -1.0)
    ranksv_parts = np.full(len_old, -1.0)
    ranksr_parts = np.full(len_old, -1.0)
    ranksp_parts = np.full(len_old, -1.0)
    pos_parts = np.full((len_old, 3), -1.0)
    vel_parts = np.full((len_old, 3), -1.0)
    hvel_parts = np.full((len_old, 3), -1.0)
    Mh_parts = np.full(len_old, -1.0)
    Np_parts = np.full(len_old, -1.0)
    downsample_parts = np.full(len_old, -1.0)
    idh_parts = np.full(len_old, -1)
    deltach_parts = np.full(len_old, -1.0)
    fenvh_parts = np.full(len_old, -1.0)

    print("compiling particle subsamples")
    start_tracker = 0
    for j in np.arange(len(halos)):
        if j % 10000 == 0:
            print("halo id", j, end = '\r')
        if mask_halos[j]:
            # updating the mask tagging the particles we want to preserve
            subsample_factor = subsample_particles(halos['N'][j] * Mpart, MT)
            submask = np.random.binomial(n = 1, p = subsample_factor, size = halos_pnum[j])
            # updating the particles' masks, downsample factors, halo mass
            mask_parts[halos_pstart[j]: halos_pstart[j] + halos_pnum[j]] = submask
            # print(j, halos_pstart, halos_pnum, p_halos, downsample_parts)
            downsample_parts[halos_pstart[j]: halos_pstart[j] + halos_pnum[j]] = p_halos[j]
            hvel_parts[halos_pstart[j]: halos_pstart[j] + halos_pnum[j]] = halos['v_L2com'][j]
            Mh_parts[halos_pstart[j]: halos_pstart[j] + halos_pnum[j]] = halos['N'][j] * Mpart # in msun / h
            Np_parts[halos_pstart[j]: halos_pstart[j] + halos_pnum[j]] = np.sum(submask)
            idh_parts[halos_pstart[j]: halos_pstart[j] + halos_pnum[j]] = halos['id'][j]
            deltach_parts[halos_pstart[j]: halos_pstart[j] + halos_pnum[j]] = halos['deltac_rank'][j]
            fenvh_parts[halos_pstart[j]: halos_pstart[j] + halos_pnum[j]] = halos['fenv_rank'][j]

            # updating the pstart, pnum, for the halos
            halos_pstart_new[j] = start_tracker
            halos_pnum_new[j] = np.sum(submask)
            start_tracker += np.sum(submask)

            if want_ranks:
                if np.sum(submask) == 0:
                    continue
                # extract particle index
                indices_parts = np.arange(
                    halos_pstart[j], halos_pstart[j] + halos_pnum[j])[submask.astype(bool)]
                indices_parts = indices_parts.astype(int)
                if np.sum(submask) == 1:
                    ranks_parts[indices_parts] = 0
                    ranksv_parts[indices_parts] = 0
                    ranksp_parts[indices_parts] = 0
                    ranksr_parts[indices_parts] = 0
                    continue
                
                # make the rankings
                theseparts = parts[
                    halos_pstart[j]: halos_pstart[j] + halos_pnum[j]][submask.astype(bool)]
                theseparts_pos = theseparts['pos']
                theseparts_vel = theseparts['vel']
                theseparts_halo_pos = halos['x_L2com'][j]
                theseparts_halo_vel = halos['v_L2com'][j]

                dist2_rel = np.sum((theseparts_pos - theseparts_halo_pos)**2, axis = 1)
                newranks = dist2_rel.argsort().argsort() 
                ranks_parts[indices_parts] = (newranks - np.mean(newranks)) / np.mean(newranks)

                v2_rel = np.sum((theseparts_vel - theseparts_halo_vel)**2, axis = 1)
                newranksv = v2_rel.argsort().argsort() 
                ranksv_parts[indices_parts] = (newranksv - np.mean(newranksv)) / np.mean(newranksv)

                # get rps
                # calc relative positions
                r_rel = theseparts_pos - theseparts_halo_pos 
                r0 = np.sqrt(np.sum(r_rel**2, axis = 1))
                r_rel_norm = r_rel/r0[:, None]

                # list of peculiar velocities of the particles
                vels_rel = theseparts_vel - theseparts_halo_vel # velocity km/s
                # relative speed to halo center squared
                v_rel2 = np.sum(vels_rel**2, axis = 1) 

                # calculate radial and tangential peculiar velocity
                vel_rad = np.sum(vels_rel*r_rel_norm, axis = 1)
                newranksr = vel_rad.argsort().argsort() 
                ranksr_parts[indices_parts] = (newranksr - np.mean(newranksr)) / np.mean(newranksr)

                # radial component
                v_rad2 = vel_rad**2 # speed
                # tangential component
                v_tan2 = v_rel2 - v_rad2

                # compute the perihelion distance for NFW profile
                m = halos['N'][j]*Mpart / h # in kg
                rs = halos['r25_L2com'][j]
                c = halos['r90_L2com'][j]/rs
                r0_kpc = r0*1000 # kpc
                alpha = 1.0/(np.log(1+c)-c/(1+c))*2*6.67e-11*m*2e30/r0_kpc/3.086e+19/1e6

                # iterate a few times to solve for rp
                x2 = v_tan2/(v_tan2+v_rad2)

                num_iters = 20 # how many iterations do we want
                factorA = v_tan2 + v_rad2
                factorB = np.log(1+r0_kpc/rs)
                for it in range(num_iters):
                    oldx = np.sqrt(x2)
                    x2 = v_tan2/(factorA + alpha*(np.log(1+oldx*r0_kpc/rs)/oldx - factorB))
                x2[np.isnan(x2)] = 1
                # final perihelion distance 
                rp2 = r0_kpc**2*x2
                newranksp = rp2.argsort().argsort() 
                ranksp_parts[indices_parts] = (newranksp - np.mean(newranksp)) / np.mean(newranksp)

        else:
            halos_pstart_new[j] = -1
            halos_pnum_new[j] = -1

    halos['npstartA'] = halos_pstart_new
    halos['npoutA'] = halos_pnum_new
    halos['randoms'] = np.random.random(len(halos)) # attaching random numbers
    halos['randoms_gaus_vrms'] = np.random.normal(loc = 0, 
        scale = halos["sigmav3d_L2com"]/np.sqrt(3), size = len(halos)) # attaching random numbers

    # output halo file 
    print("outputting new halo file ")
    # output_dir = savedir+'/halos_xcom_'+str(i)+'_seed'+str(newseed)+'_abacushodMT_new.h5'
    if os.path.exists(outfilename_halos):
        os.remove(outfilename_halos)
    print(outfilename_halos, outfilename_particles)
    newfile = h5py.File(outfilename_halos, 'w')
    dataset = newfile.create_dataset('halos', data = halos[mask_halos])
    newfile.close()

    # output the new particle file
    print("adding rank fields to particle data ")
    mask_parts = mask_parts.astype(bool)
    parts = parts[mask_parts]
    print("pre process particle number ", len_old, " post process particle number ", len(parts))
    if want_ranks:
        parts['ranks'] = ranks_parts[mask_parts]
        parts['ranksv'] = ranksv_parts[mask_parts]
        parts['ranksr'] = ranksr_parts[mask_parts]
        parts['ranksp'] = ranksp_parts[mask_parts]
    parts['downsample_halo'] = downsample_parts[mask_parts]
    parts['halo_vel'] = hvel_parts[mask_parts]
    parts['halo_mass'] = Mh_parts[mask_parts]
    parts['Np'] = Np_parts[mask_parts]
    parts['halo_id'] = idh_parts[mask_parts]
    parts['randoms'] = np.random.random(len(parts))
    parts['halo_deltac'] = deltach_parts[mask_parts]
    parts['halo_fenv'] = fenvh_parts[mask_parts]

    print("are there any negative particle values? ", np.sum(parts['downsample_halo'] < 0), 
        np.sum(parts['halo_mass'] < 0))
    print("outputting new particle file ")
    # output_dir = savedir+'/particles_xcom_'+str(i)+'_seed'+str(newseed)+'_abacushodMT_new.h5'
    if os.path.exists(outfilename_particles):
        os.remove(outfilename_particles)
    newfile = h5py.File(outfilename_particles, 'w')
    dataset = newfile.create_dataset('particles', data = parts)
    newfile.close()

    print("pre process particle number ", len_old, " post process particle number ", len(parts))

def main(path2config, params = None, alt_simname = None, newseed = 600):
    print("compiling compaso halo catalogs into subsampled catalogs")

    config = yaml.safe_load(open(path2config))
    # update params if needed
    if params:
        config.update(params)
    if alt_simname: 
        config['sim_params']['sim_name'] = alt_simname

    simname = config['sim_params']['sim_name'] # "AbacusSummit_base_c000_ph006"
    simdir = config['sim_params']['sim_dir']
    z_mock = config['sim_params']['z_mock']
    savedir = config['sim_params']['subsample_dir']+simname+"/z"+str(z_mock).ljust(5, '0') 
    cleaning = config['sim_params']['cleaned_halos']

    halo_info_fns = \
    list((Path(simdir) / Path(simname) / 'halos' / ('z%4.3f'%z_mock) / 'halo_info').glob('*.asdf'))
    numslabs = len(halo_info_fns)

    tracer_flags = config['HOD_params']['tracer_flags']
    MT = False
    if tracer_flags['ELG'] or tracer_flags['QSO']:
        MT = True
    want_ranks = config['HOD_params']['want_ranks']
    want_AB = config['HOD_params']['want_AB']
    # N_dim = config['HOD_params']['Ndim']

    os.makedirs(savedir, exist_ok = True)
    
    p = multiprocessing.Pool(config['prepare_sim']['Nparallel_load'])
    p.starmap(prepare_slab, zip(range(numslabs), repeat(savedir), 
        repeat(simdir), repeat(simname), repeat(z_mock), 
        repeat(tracer_flags), repeat(MT), repeat(want_ranks), 
        repeat(want_AB), repeat(cleaning), repeat(newseed)))
    p.close()
    p.join()

    # print("done, took time ", time.time() - start)

class ArgParseFormatter(argparse.RawDescriptionHelpFormatter, argparse.ArgumentDefaultsHelpFormatter):
    pass

if __name__ == "__main__":

    # parsing arguments
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=ArgParseFormatter)
    parser.add_argument('--path2config', help='Path to the config file', default=DEFAULTS['path2config'])
    parser.add_argument('--alt_simname',
                        help='alternative simname to process, like "AbacusSummit_base_c000_ph003"',
                       )
    parser.add_argument('--newseed',
                        help='alternative random number seed, positive integer', default = 600
                       )
    args = vars(parser.parse_args())

    main(**args)

    print("done")
    # # run a series of simulations
    # param_dict = {
    # 'sim_params' :
    #     {
    #     'sim_name': 'AbacusSummit_base_c000_ph006',                                 # which simulation 
    #     # 'sim_dir': '/mnt/gosling2/bigsims/',                                        # where is the simulation
    #     'sim_dir': '/mnt/marvin1/syuan/scratch/bigsims/',                                        # where is the simulation
    #     'output_dir': '/mnt/marvin1/syuan/scratch/data_mocks_georgios',          # where to output galaxy mocks
    #     'subsample_dir': '/mnt/marvin1/syuan/scratch/data_summit/',                 # where to output subsample data
    #     'z_mock': 0.5,                                                             # which redshift slice
    #     'Nthread_load': 7,                                                          # number of thread for organizing simulation outputs (prepare_sim)
    #     'cleaned_halos': False
    #     }
    # }
    # for i in range(25):
    #     param_dict['sim_params']['sim_name'] = 'AbacusSummit_base_c000_ph'+str(i).zfill(3)
    #     main(**args, params = param_dict)

    # other_cosmologies = [
    # 'AbacusSummit_base_c100_ph000',
    # 'AbacusSummit_base_c101_ph000',
    # 'AbacusSummit_base_c102_ph000',
    # 'AbacusSummit_base_c103_ph000',
    # 'AbacusSummit_base_c112_ph000',
    # 'AbacusSummit_base_c113_ph000',
    # 'AbacusSummit_base_c104_ph000',
    # 'AbacusSummit_base_c105_ph000',
    # 'AbacusSummit_base_c108_ph000',
    # 'AbacusSummit_base_c109_ph000',
    # 'AbacusSummit_base_c009_ph000'
    # ]
 
    # for ecosmo in other_cosmologies:
    #     print(ecosmo)
    #     param_dict['sim_params']['sim_name'] = ecosmo
    #     main(**args, params = param_dict)



#!/usr/bin/python
from __future__ import print_function
import json, logging, os, random, re, signal, socket, subprocess as SUB, sys, time

from multiprocessing import Process as mpProcess, Queue as mpQueue
try:
    from queue import Queue, Empty
except ImportError:
    from Queue import Queue, Empty
from threading import BoundedSemaphore, Thread, Lock
from ast import literal_eval

DisBatchPath, ImportDir, PathsFixed = None, None, False # <= May need to set these, setting PathsFixed to True as well.

# Handle self-modification invocation early, before messing with paths and other imports
if '__main__' == __name__ and sys.argv[1:] == ["--fix-paths"]:
    import tempfile
    DisBatchPath = os.path.realpath(__file__)
    if not os.path.exists(DisBatchPath):
        print('Unable to find myself; set DisBatchPath and ImportDir manually at the top of disBatch.py.', file=sys.stderr)
        sys.exit(1)
    DisBatchDir = os.path.dirname(DisBatchPath)
    with open(DisBatchPath, 'r') as fi:
        with tempfile.NamedTemporaryFile('w', prefix='disBatch.py.', dir=DisBatchDir, delete=False) as fo:
            found = False
            for l in fi:
                if l.startswith('DisBatchPath, ImportDir, PathsFixed ='):
                    assert not found
                    found = True
                    l = 'DisBatchPath, ImportDir, PathsFixed = %r, %r, True\n'%(DisBatchPath, DisBatchDir)
                    print("Changing path info to %r"%l, file=sys.stderr)
                fo.write(l)
            assert found
            os.fchmod(fo.fileno(), os.fstat(fi.fileno()).st_mode)
    os.rename(DisBatchPath, DisBatchPath+'.prev')
    os.rename(fo.name, DisBatchPath)
    sys.exit(0)

if not PathsFixed:
    # Try to guess
    DisBatchPath = os.path.realpath(__file__)
    ImportDir = os.path.dirname(DisBatchPath)

PythonPath = os.getenv('PYTHONPATH', '')
if ImportDir:
    # to find kvsstcp:
    sys.path.append(ImportDir)
    # for subprocesses:
    PythonPath = PythonPath + ':' + ImportDir if PythonPath else ImportDir
    os.environ['PYTHONPATH'] = PythonPath

try:
    import kvsstcp
except ImportError:
    if PathsFixed:
        print('This script is looking in the wrong place for "kvssctp". Try running "%s --fix-paths" or editing it by hand.'%DisBatchPath, file=sys.stderr)
    else:
        print('''
Could not find kvsstcp. If there is a "kvsstcp" directory in "%s",
try running "%s --fix-paths". Otherwise review the installation
instructions.
'''%(ImportDir, DisBatchPath), file=sys.stderr)
    sys.exit(1)
    
myHostname = socket.gethostname()
myPid = os.getpid()

# Note that even though these are case insensitive, only lines that start with upper-case '#DISBATCH' prefixes are tested
dbbarrier = re.compile('^#DISBATCH BARRIER(?: (.+)?)?$', re.I)
dbcomment = re.compile('^\s*(#|$)')
dbprefix  = re.compile('^#DISBATCH PREFIX (.*)$', re.I)
dbrepeat  = re.compile('^#DISBATCH REPEAT\s+(?P<repeat>[0-9]+)(?:\s+start\s+(?P<start>[0-9]+))?(?:\s+step\s+(?P<step>[0-9]+))?(?: (?P<command>.+))?\s*$', re.I)
dbsuffix  = re.compile('^#DISBATCH SUFFIX (.*)$', re.I)
dbpernode = re.compile('^#DISBATCH PERNODE (.*)$', re.I)

# Special ID for "out of band" task events
TaskIdOOB = -1
CmdPoison = '!!Poison!!'

def compHostnames(h0, h1):
    return h0.split('.', 1)[0] == h1.split('.', 1)[0]

def logfile(context, suffix=''):
    '''Standardized file path construction for log files.'''
    f = context.name
    try:
        f += "_%s"%context.node
    except AttributeError:
        pass
    if suffix:
        f += "_%s"%suffix
    return f

def waitTimeout(sub, timeout, interval=1):
    r = sub.poll()
    while r is None and timeout > 0:
        time.sleep(interval)
        timeout -= interval
        r = sub.poll()
    return r

def killPatiently(sub, name, timeout=15):
    if not sub: return
    r = sub.poll()
    if r is None:
        logger.info('Waiting for %s to finish...', name)
        r = waitTimeout(sub, timeout)
    if r is None:
        logger.warn('Terminating %s...', name)
        try:
            sub.terminate()
        except OSError:
            pass
        r = waitTimeout(sub, timeout)
    if r is None:
        logger.warn('Killing %s.', name)
        try:
            sub.kill()
        except OSError:
            pass
        r = sub.wait()
    if r:
        logger.info("%s returned %d", name, r)
    return r

class BatchContext(object):
    def __init__(self, sysid, jobid, nodes, cylinders):
        self.sysid, self.jobid, self.nodes, self.cylinders = sysid, jobid, nodes, cylinders
        self.wd = os.getcwd() #TODO: Easy enough to override, but still... Is this the right place for this?
        self.error = False # engine errors (non-zero return values)
        self.retireCmd = None

    def __str__(self):
        return 'Batch system: %s\nJobID: %s\nNodes: %r\nCylinders: %r\n'%(self.sysid, self.jobid, self.nodes, self.cylinders)

    def launch(self, kvs):
        '''Launch the engine processes on all the nodes by calling launchNode for each.'''
        kvs.put('.context', self)
        self.engines = dict() # live subprocesses
        for n in self.nodes:
            self.engines[n] = self.launchNode(n)

    def poll(self):
        '''Check if any engines have stopped.'''
        for n, e in list(self.engines.items()):
            r = e.poll()
            if r is not None:
                logger.info('Engine %s exited: %d', n, r)
                del self.engines[n]
                self.retireNode(n, r)

    def launchNode(self, node):
        '''Launch an engine for a single node.  Should return a subprocess handle (unless launch itself is overridden).'''
        raise NotImplementedError('%s.launchNode is not implemented' % type(self))

    def retireEnv(self, node, ret):
        '''Generate an environment for running the retirement command for a given node.'''
        env = os.environ.copy()
        env['NODE'] = node
        env['RETCODE'] = str(ret)
        env['ACTIVE'] = ','.join(self.engines.keys())
        env['RETIRED'] = ','.join(set(self.nodes).difference(self.engines))
        return env

    def retireNode(self, node, ret):
        '''Called when a node has exited.  May be overridden to release resources.'''
        if ret: self.error = True
        if self.retireCmd:
            logger.info('Retiring node "%s" with command %s', node, str(self.retireCmd))
            env = self.retireEnv(node, ret)
            try:
                SUB.check_call(self.retireCmd, close_fds=True, shell=True, env=env)
            except Exception as e:
                logger.warn('Retirement planning needs improvement: %s', repr(e))
        else:
            logger.info('Retiring node "%s" (no-op)', node)

    def finish(self):
        '''Check that all engines completed successfully and return True on success.'''
        for n, e in self.engines.items():
            r = killPatiently(e, 'engine ' + n)
            if r: self.error = True # also handled by retireNode
        return not self.error

    def setNode(self, node=None):
        '''Try to determine the hostname of this engine from the pov of the launcher.'''
        # This is just a fallback. Implementations should try to determine node as appropriate.
        # Could just default to node=myHostname, but then we lose special domain-name matching
        if not node:
            for n in self.nodes:
                if compHostnames(n, myHostname):
                    node = n
                    break
        self.node = node
        try:
            self.nodeId = self.nodes.index(self.node)
        except ValueError:
            # Should we instead assume 0 or carry on with none?
            raise LookupError('Couldn\'t find nodeId for %s in "%s".' % (node or myHostname, self.nodes))

# Convert nodelist format (slurm specific?) to an expanded list of nodes.
#    nl     => hosts[,nl]
#    hosts  => prefix[\[ranges\]]
#    ranges => range[,ranges]
#    range  => lo[-hi]
# where lo and hi are numbers
def nl2flat(nl):
    return SUB.check_output(["scontrol", "show", "hostnames", nl], universal_newlines=True).splitlines()

class SlurmContext(BatchContext):
    def __init__(self):
        jobid = os.environ['SLURM_JOBID']
        nodes = nl2flat(os.environ['SLURM_NODELIST'])

        cylinders = []
        for tr in os.environ['SLURM_TASKS_PER_NODE'].split(','):
            m = re.match(r'([^\(]+)(?:\(x([^\)]+)\))?', tr)
            c, m = m.groups()
            if m == None: m = '1'
            cylinders += [int(c)]*int(m)

        super(SlurmContext, self).__init__('SLURM', jobid, nodes, cylinders)
        self.driverNode = None
        self.retireCmd = "scontrol update JobId=\"$SLURM_JOBID\" NodeList=\"${DRIVER_NODE:+$DRIVER_NODE,}$ACTIVE\""

    def launchNode(self, n):
        return SUB.Popen(['srun', '-N', '1', '-n', '1', '-w', n, DisBatchPath, '--engine', '-n', n, kvsserver], close_fds=True)

    def retireEnv(self, node, ret):
        env = super(SlurmContext, self).retireEnv(node, ret)
        if self.driverNode:
            env['DRIVER_NODE'] = self.driverNode
        return env

    def retireNode(self, node, ret):
        if compHostnames(node, myHostname):
            self.driverNode = node
        super(SlurmContext, self).retireNode(node, ret)

    def setNode(self, node=None):
        super(SlurmContext, self).setNode(node or os.getenv('SLURMD_NODENAME'))

#TODO:
#class GEContext(BatchContext):
#class LSFContext(BatchContext):
#class PBSContext(BatchContext):

# The ssh context should be generally applicable when all else fails
# (or there is no resource manager).
#
# To use, set the environment variable DISBATCH_SSH_NODELIST. E.g.:
#     DISBATCH_SSH_NODELIST=hostname0:4,hostname1:5
# indicates 4 cylinders (execution entities) should run on hostname0
# and 5 on hostname1.
#
# You can also specify a "job id" via DISBATCH_SSH_JOBID. If you do
# not provide one, one will be created from the PID and epoch time.
class SSHContext(BatchContext):
    def __init__(self, nodelist=os.getenv('DISBATCH_SSH_NODELIST')):
        jobid = os.getenv('DISBATCH_SSH_JOBID', '%d_%.6f'%(myPid, time.time()))

        cylinders, nodes = [], []
        if type(nodelist) is not str: nodelist = ','.join(nodelist)
        for p in nodelist.split(','):
            p = p.strip()
            if not p: continue
            try:
                n, e = p.rsplit(':', 1)
                e = int(e)
            except ValueError:
                raise ValueError('SSH nodelist items must be HOST:COUNT')
            if n == 'localhost': n = myHostname
            nodes.append(n)
            cylinders.append(e)

        super(SSHContext, self).__init__('SSH', jobid, nodes, cylinders)

    def launchNode(self, n):
        prefix = [] if compHostnames(n, myHostname) else ['ssh', n, 'PYTHONPATH=' + PythonPath]
        return SUB.Popen(prefix + [DisBatchPath, '--engine', '-n', n, kvsserver], stdin=open(os.devnull, 'r'), stdout=open(logfile(self, '%s_engine_wrap.out'%n), 'w'), stderr=open(logfile(self, '%s_engine_wrap.err'%n), 'w'), close_fds=True)

def probeContext():
    if 'SLURM_JOBID' in os.environ: return SlurmContext()
    #if ...: return GEContext()
    #if ...: LSFContext()
    #if ...: PBSContext()
    if 'DISBATCH_SSH_NODELIST' in os.environ: return SSHContext()


class TaskInfo(object):
    def __init__(self, taskId, taskStreamIndex, taskRepIndex, taskCmd, taskKey, host = '', pid = 0, returncode = 0, start = 0, end = 0, outbytes = 0, outdata = '', errbytes = 0, errdata = '', skip = False):
        self.taskId, self.taskStreamIndex, self.taskRepIndex, self.taskCmd, self.taskKey = taskId, taskStreamIndex, taskRepIndex, taskCmd, taskKey
        self.host, self.pid, self.returncode, self.start, self.end, self.outbytes, self.outdata, self.errbytes, self.errdata = host, pid, returncode, start, end, outbytes, outdata, errbytes, errdata
        self.skip = skip

    def flags(self):
        return (  ('R' if self.returncode else ' ')
                + ('O' if self.outbytes   else ' ')
                + ('E' if self.errbytes   else ' ')
                + ('S' if self.skip       else ''))

    def __str__(self):
        # If this changes, update parseStatusFile below and disBatcher.py too
        return '\t'.join([str(x) for x in [self.flags(), self.taskId, self.taskStreamIndex, self.taskRepIndex, self.host, self.pid, self.returncode, self.end - self.start, self.start, self.end, self.outbytes, repr(self.outdata), self.errbytes, repr(self.errdata), repr(self.taskCmd)]])

    def __eq__(self, other):
        return type(self) is type(other) and self.taskId == other.taskId and self.taskStreamIndex == other.taskStreamIndex and self.taskRepIndex == other.taskRepIndex and self.taskCmd == other.taskCmd # and self.taskKey == other.taskKey

    def __ne__(self, other):
        return not self == other

class BarrierTask(TaskInfo):
    def __init__(self, taskId, taskStreamIndex, taskRepIndex, taskCmd, key=None, check=False, returncode=0, errdata=''):
        super(BarrierTask, self).__init__(taskId, taskStreamIndex, taskRepIndex, taskCmd, key, myHostname, myPid, returncode=returncode, errdata=errdata)
        self.check = check

    def flags(self):
        return 'B'

class DoneTask(BarrierTask):
    '''Implicit barrier posted when there are no more tasks.'''
    def __init__(self, taskId, taskStreamIndex, err=''):
        super(DoneTask, self).__init__(taskId, taskStreamIndex, -1, '', 'done!', returncode=1 if err else 0, errdata=err)

    def flags(self):
        return 'D'

def parseStatusFiles(*files):
    status = dict()
    for f in files:
        with open(f, 'r') as s:
            for l in s:
                d = l.split('\t')
                if len(d) != 15:
                    logger.warn('Invalid status line (ignoring): %r'%l)
                    continue
                if d[0] in 'BD': continue
                ti = TaskInfo(int(d[1]), int(d[2]), int(d[3]), literal_eval(d[14]), '.task', d[4], int(d[5]), int(d[6]), float(d[8]), float(d[9]), int(d[10]), literal_eval(d[11]), int(d[12]), literal_eval(d[13]), True)
                try:
                    # successful tasks take precedence
                    if status[ti.taskId].returncode <= ti.returncode: continue
                except KeyError:
                    pass
                status[ti.taskId] = ti
    return status

##################################################################### DRIVER

# When the user specifies tasks will be passed through a KVS, this
# class generates an interable that feeds task from the KVS.
class KVSTaskSource(object):
    def __init__(self, kvs):
        self.kvs = kvs.clone()
        self.name = self.kvs.get('task source name', False)
        self.taskkey = self.name + ' task'
        self.resultkey = self.name + ' result %d'
        self.donetask = self.name + ' done!'

    def __next__(self):
        t = self.kvs.get(self.taskkey, False)
        if t == self.donetask:
            self.kvs.close()
            raise StopIteration
        return t

    def done(self):
        kvs = self.kvs.clone()
        kvs.put(self.taskkey, self.donetask)
        kvs.close()

# When the user specifies a command that will be generating tasks,
# this class wraps the command's execution so we can trigger a
# shutdown if the user's command fails to send an indication that task
# generation is done.
class TaskProcess():
    def __init__(self, taskSource, command, **kwargs):
        self.taskSource = taskSource
        self.command = command
        self.p = SUB.Popen(command, **kwargs)
        self.r = None

    def poll(self):
        if self.r is not None: return
        self.r = self.p.poll()
        if self.r is None: return
        # TODO: send done on success
        logger.info('Task generating command has exited: %s %d.', repr(self.command), self.r)
        # post a done just in case the process didn't
        self.taskSource.done()

# Given a task source (generating task command lines), parse the lines and
# produce a TaskInfo generator.
def taskGenerator(tasks, context):
    tsx = 0 # "line number" of current task
    taskCounter = 0 # next taskId
    prefix = suffix = ''

    while 1:
        tsx += 1
        try:
            t = next(tasks)
        except StopIteration:
            # Signals there will be no more tasks.
            break

        logger.debug('Task: %s', t)

        # Split on newlines.
        #
        # This allows tasks submitted through kvs with or without newlines,
        # including multiple tasks per item, or from files (always with single
        # trailing newline).
        #
        # Note that multiple lines in the same item get the same streamIndex,
        # but this shouldn't be a problem.  (Alternatively could increment tsx
        # inside this loop instead.)
        for t in t.splitlines():

            if t.startswith('#DISBATCH '):
                m = dbprefix.match(t)
                if m:
                    prefix = m.group(1)
                    continue
                m = dbsuffix.match(t)
                if m:
                    suffix = m.group(1)
                    continue
                m = dbrepeat.match(t)
                if m:
                    repeats, rx, step = int(m.group('repeat')), 0, 1
                    g = m.group('start')
                    if g: rx = int(g)
                    g = m.group('step')
                    if g: step = int(g)
                    logger.info('Processing repeat: %d %d %d', repeats, rx, step)
                    cmd = prefix + (m.group('command') or '') + suffix
                    while repeats > 0:
                        yield TaskInfo(taskCounter, tsx, rx, cmd, '.task')
                        taskCounter += 1
                        rx += step
                        repeats -= 1
                    continue
                m = dbpernode.match(t)
                if m:
                    cmd = m.group(1)
                    for rx, node in enumerate(context.nodes):
                        yield TaskInfo(taskCounter, tsx, rx, prefix + cmd + suffix, '.node.'+node)
                        taskCounter += 1
                    continue
                m = dbbarrier.match(t)
                if m:
                    bkey = m.group(1)
                    if bkey == 'CHECK':
                        check = True
                        bkey = None
                    else:
                        check = False
                    yield BarrierTask(taskCounter, tsx, -1, t, key=bkey, check=check)
                    taskCounter += 1
                    continue
                logger.error('Unknown #DISBATCH directive: %s', t)

            if dbcomment.match(t):
                # Comment or empty line, ignore
                continue

            yield TaskInfo(taskCounter, tsx, -1, prefix + t + suffix, '.task')
            taskCounter += 1

    logger.info('Processed %d tasks.', taskCounter)
    yield DoneTask(taskCounter, tsx)

def statusTaskFilter(tasks, status, retry=False, force=False):
    while True:
        t = next(tasks)
        s = status.get(t.taskId)
        if s and (not retry or s.returncode == 0):
            # skip
            if s != t:
                msg = 'Recovery status file task mismatch %s:\n' + str(s) + '\n' + str(t)
                if force:
                    logger.warn(msg, '-- proceeding anyway')
                else:
                    raise Exception(msg % '(use --force-resume to proceed anyway)')
            yield s
        else:
            yield t

# Main control loop that sends new tasks to the execution engines.
class Feeder(Thread):
    def __init__(self, kvs, context, tasks, taskSlots):
        super(Feeder, self).__init__(name='Feeder')
        self.daemon = True
        self.context = context
        self.taskGenerator = tasks
        self.taskSlots = taskSlots
        self.kvs = kvs.clone()
        # just used for status reporting (not thread safe):
        self.done = False
        self.barrier = None
        self.start()

    def run(self):
        try:
            self.main()
        except Exception as e:
            logger.error('Feeder error: %s', e)
            self.kvs.put('.finished task', DoneTask(-1, 0, str(e)))
            raise

    def main(self):
        self.kvs.put('.common env', {'DISBATCH_JOBID': str(self.context.jobid), 'DISBATCH_NAMETASKS': self.context.name}) #TODO: Add more later?
        while True:
            # Wait for a slot
            self.taskSlots.acquire()

            if self.done:
                tinfo = DoneTask(-1, 0, 'aborted')
            else:
                # Request the next task
                tinfo = next(self.taskGenerator)

            if tinfo.skip:
                self.kvs.put('.finished task', tinfo)
                continue

            if isinstance(tinfo, BarrierTask):
                if isinstance(tinfo, DoneTask):
                    self.done = True
                    # Post the poison pill. This may trigger retirement of engines.
                    self.kvs.put('.task', [TaskIdOOB, -1, -1, CmdPoison])

                self.barrier = tinfo.taskKey or True
                # Enter a barrier. We'll exit when all tasks
                # issued to this point have completed.
                tinfo.start = time.time()
                logger.info('Entering barrier (key is %s).', repr(tinfo.taskKey))

                # Wait for all the other task slots to be free (_initial_value is totalSlots)
                # These ar ethen released by Driver when barrier is finished
                for i in range(self.taskSlots._initial_value-1):
                    self.taskSlots.acquire()

                # Completed all tasks up to, but not including the barrier. Now complete the barrier.
                logger.info('Finishing barrier.')
                # post a task finished for the current barrier, just like any other task, triggering the release of all taskSlots
                tinfo.end = time.time()
                self.kvs.put('.finished task', tinfo)
                self.barrier = None

                if isinstance(tinfo, DoneTask): break
                continue

            # At this point, we have a task
            tpayload = [tinfo.taskId, tinfo.taskStreamIndex, tinfo.taskRepIndex, tinfo.taskCmd]
            logger.info('Posting task: %r', tpayload)
            self.kvs.put(tinfo.taskKey, tpayload)

        self.kvs.close()

# Main control loop that processes completed tasks.
class Driver(Thread):
    def __init__(self, kvs, context, tasks, trackResults=None, mailTo=None, mailFreq=1):
        super(Driver, self).__init__(name='Driver')
        self.context = context
        self.kvs = kvs.clone()
        self.mailTo = mailTo
        self.mailFreq = mailFreq

        self.trackResults = trackResults
        self.taskSlots = BoundedSemaphore(sum(self.context.cylinders))
        self.feeder = Feeder(self.kvs, context, tasks, self.taskSlots)

        self.finished = 0
        self.failed = 0

        self.statusFile = open(logfile(self.context, 'status.txt'), 'w+')
        self.statusLastOffset = self.statusFile.tell()

        self.daemon = True
        self.start()

    def sendNotification(self):
        try:
            import smtplib
            from email.mime.text import MIMEText
            self.statusFile.seek(self.statusLastOffset)
            msg = MIMEText('Last %d:\n\n'%self.mailFreq + self.statusFile.read())
            msg['Subject'] = '%s (%s) has completed %d tasks'%(self.context.name, self.context.jobid, self.finished)
            if self.failed:
                msg['Subject'] += ' (%d failed)'%self.failed
            msg['From'] = self.mailTo
            msg['To'] = self.mailTo
            s = smtplib.SMTP()
            s.connect()
            s.sendmail([self.mailTo], [self.mailTo], msg.as_string())
            self.statusLastOffset = statusfo.tell()
        except Exception as e:
            logger.warn('Failed to send notification message: "%s". Disabling.', e)
            self.mailTo = None
            # Be sure to seek back to EOF to append
            self.statusFile.seek(0, 2)

    def updateStatus(self):
        active = self.taskSlots._initial_value
        try:
            active -= self.taskSlots._value
        except AttributeError:
            active -= self.taskSlots._Semaphore__value
        status = dict(more = not self.feeder.done, barrier = self.feeder.barrier,
                finished = self.finished, failed = self.failed,
                # base active on semaphore value (correct except for barriers)
                active = active)
        # Make changes visible via KVS.
        logger.debug('Posting status: %r', status)
        self.kvs.get('DisBatch status', False)
        self.kvs.put('DisBatch status', json.dumps(status, default=repr), b'JSON')

    def run(self):
        self.kvs.put('DisBatch status', '<Starting...>', False)
        while 1:
            logger.debug('Driver loop: %d', self.finished)

            # Wait for a finished task
            tinfo = self.kvs.get('.finished task')
            logger.debug('Finished task: %s', tinfo)

            if isinstance(tinfo, BarrierTask):
                if isinstance(tinfo, DoneTask): break
                # Complete the barrier task itself, exit barrier mode.
                logger.info('Finished barrier.')
                if tinfo.check and self.failed:
                    # a "check" barrier fails if any tasks before it do
                    tinfo.returncode = 1
                    # stop the feeder (prompting DoneTask)
                    self.feeder.done = True
                # If user specified a KVS key, use it to signal the barrier is done.
                elif tinfo.taskKey:
                    logger.info('put %s: %d.', tinfo.taskKey, tinfo.taskId)
                    self.kvs.put(tinfo.taskKey, str(tinfo.taskId), False)
                # Release the rest of the slots (_initial_value is totalSlots)
                for i in range(self.taskSlots._initial_value-1):
                    self.taskSlots.release()

            self.taskSlots.release()
            self.finished += 1
            if tinfo.returncode: self.failed += 1
            self.statusFile.write(str(tinfo)+'\n')
            self.statusFile.flush()

            # Maybe we want to track results by streamIndex instead of taskId?  But then there could be more than one per key.
            if self.trackResults: self.kvs.put(self.trackResults%tinfo.taskId, str(tinfo), False)
            if self.mailTo and self.finished%self.mailFreq == 0:
                self.sendNotification()

            # Make changes visible via KVS.
            self.updateStatus()

        logger.info('Driver done')
        self.statusFile.close()
        self.kvs.close()
        self.feeder.join()

##################################################################### ENGINE

# A simple class to count the number of bytes from a file stream (e.g., pipe),
# and possibly collect the first and/or last few bytes of it
class OutputCollector(Thread):
    def __init__(self, pipe, takeStart=0, takeEnd=0):
        super(OutputCollector, self).__init__(name='OutputCollector')
        # We don't really care for python's file abstraction -- get back a real fd
        self.pipefd = os.dup(pipe.fileno())
        pipe.close()
        self.takeStart = takeStart
        self.takeEnd = takeEnd
        self.dataStart = b''
        self.dataEnd = b''
        self.bytes = 0
        self.daemon = True
        self.start()

    def read(self, count):
        try:
            return os.read(self.pipefd, count)
        except OSError:
            return ''

    def run(self):
        start = self.takeStart
        end = self.takeEnd
        while 1:
            if start > 0:
                r = self.read(start)
                self.dataStart += r
                start -= len(r)
            else:
                r = self.read(4096)
                if end > 0:
                    if len(r) >= end:
                        self.dataEnd = r[-end:]
                    else:
                        self.dataEnd = self.dataEnd[-end+len(r):] + r
            if not r: return
            self.bytes += len(r)

    def stop(self):
        self.join(5)
        try:
            # In case someone still has the pipe open, close our end to force this thread to stop
            os.close(self.pipefd)
        except OSError:
            pass

    def __str__(self):
        s = self.dataStart
        if self.dataEnd:
            s += b'...' + self.dataEnd
        if type(s) is not str:
            # for python3... would be better to keep raw bytes
            s = s.decode('utf-8', 'ignore')
        return s

# Once we know the nodes participating in the run, we start an engine
# on each node. The engine in turn starts the number of cylinders
# (execution entities) specified for the node (a map of nodes to
# cylinder count is conveyed via the KVS). Each cylinder executes one
# task at a time. The Injector waits for an available cylinder and
# then waits for a task (grabbing a task before a cylinder is ready
# would prevent an idle cylinder of another engine from performing the
# task). The task is passed to the run method via a queue. The run
# method uses queues to feed tasks to the cylinders and accept
# results from them. It tracks work in progress and notes when the
# "poison" task has been received. This indicates no more tasks will
# be coming. At this point, once all work in progress is done, the
# engine can retire.
class EngineBlock(Thread):
    class Cylinder(mpProcess):
        def __init__(self, context, env, envres, ciq, coq, cylinderId):
            super(EngineBlock.Cylinder, self).__init__()
            self.daemon = True
            self.context, self.ciq, self.coq, self.cylinderId = context, ciq, coq, cylinderId
            self.localEnv = env.copy()
            for v, l in envres.items():
                try:
                    self.localEnv[v] = l[cylinderId]
                except IndexError:
                    # safer to set it empty than delete it for most cases
                    self.localEnv[v] = ''
            self.taskProc = None
            self.start()

        def run(self):
            signal.signal(signal.SIGTERM, lambda s, f: sys.exit(1))
            try:
                self.main()
            finally:
                logger.info('Cylinder %d stopping.', self.cylinderId)
                self.coq.put(('done', 'stopped', False))
                killPatiently(self.taskProc, 'cylinder %d subproc' % self.cylinderId, 2)

        def main(self):
            self.pgid = os.getpgid(0)
            logger.info('Cylinder %d firing, %d, %d.', self.cylinderId, self.pid, self.pgid)
            while 1:
                (taskId, taskStreamIndex, taskRepIndex, taskCmd), throttled = self.ciq.get()
                if taskId == TaskIdOOB:
                    break
                logger.info('Cylinder %d executing %s.', self.cylinderId, repr([taskId, taskStreamIndex, taskRepIndex, taskCmd]))
                self.localEnv['DISBATCH_STREAM_INDEX'], self.localEnv['DISBATCH_REPEAT_INDEX'], self.localEnv['DISBATCH_TASKID'] = str(taskStreamIndex), str(taskRepIndex), str(taskId)
                t0 = time.time()
                try:
                    self.taskProc = SUB.Popen(['/bin/bash', '-c', taskCmd], env=self.localEnv, stdin=None, stdout=SUB.PIPE, stderr=SUB.PIPE, preexec_fn=os.setsid, close_fds=True)
                    pid = self.taskProc.pid
                    obp = OutputCollector(self.taskProc.stdout, 40, 40)
                    ebp = OutputCollector(self.taskProc.stderr, 40, 40)
                    r = self.taskProc.wait()
                    self.taskProc = None
                    t1 = time.time()
                    
                    obp.stop()
                    ebp.stop()
                    ti = TaskInfo(taskId, taskStreamIndex, taskRepIndex, taskCmd, '.finished task', self.context.node, pid, r, t0, t1, obp.bytes, str(obp), ebp.bytes, str(ebp))
                except Exception as e:
                    self.taskProc = None
                    t1 = time.time()
                    estr = str(e)
                    ti = TaskInfo(taskId, taskStreamIndex, taskRepIndex, taskCmd, '.finished task', self.context.node, -1, getattr(e, 'errno', 200), t0, t1, 0, '', len(estr), estr)

                logger.info('Cylinder %s completed: %s', self.cylinderId, ti)
                self.coq.put(('done', ti, throttled))

    class Injector(Thread):
        def __init__(self, kvs, key, throttle, fuelline):
            super(EngineBlock.Injector, self).__init__(name='Injector')
            self.daemon = True
            self.kvs = kvs.clone()
            self.key = key
            self.throttle = throttle
            self.fuelline = fuelline
            self.start()

        def run(self):
            while 1:
                if self.throttle: self.throttle.acquire()
                ti = self.kvs.get(self.key)
                logger.debug('Injector got task %s', repr(ti))
                self.fuelline.put(('task', ti, bool(self.throttle)))

    def __init__(self, kvs, context):
        super(EngineBlock, self).__init__(name='EngineBlock')
        self.daemon = True
        self.context = context
        cylinders = context.cylinders[context.nodeId]

        env = os.environ
        envres = {}
        for v in context.envres:
            e = env.get(v)
            if e:
                l = e.split(',')
                if len(l) < cylinders:
                    logger.error('Requested envres variable "%s" has too few values, decreasing cylinders to match: %s', v, e)
                    # This may not be safe: driver is still feeding tasks based on original count
                    cylinders = len(l)
                elif len(l) > cylinders:
                    logger.warning('Requested envres variable "%s" has too many values, so some resources will not be used: %s', v, e)
                envres[v] = l
            else:
                logger.warning('Requested envres variable "%s" not found', v)

        self.parent = kvs
        self.kvs = kvs.clone()
        env.update(self.kvs.view('.common env'))
        # Note we are using the Queue construct from the
        # mulitprocessing module---we need to coordinate between
        # independent processes.
        self.ciq, self.coq, self.throttle = mpQueue(), mpQueue(), BoundedSemaphore(cylinders)
        self.Injector(kvs, ".node." + context.node, None, self.coq)
        self.Injector(kvs, ".task", self.throttle, self.coq)
        self.cylinders = [self.Cylinder(context, env, envres, self.ciq, self.coq, x) for x in range(cylinders)]
        self.start()

    def run(self):
        try:
            self.main()
        finally:
            self.parent.close()

    def main(self):
        inFlight, liveCylinders = 0, len(self.cylinders)
        while liveCylinders:
            tag, o, throttled = self.coq.get()
            logger.info('Run loop: %d %d %s %s', inFlight, liveCylinders, tag, str(o))
            if tag == 'done':
                inFlight -= 1
                if throttled: self.throttle.release()
                if o == 'stopped':
                    liveCylinders -= 1
                else:
                    self.kvs.put(o.taskKey, o)
            elif tag == 'task':
                # Handle control messages (TaskIdOOB). The default at the moment
                # is to put it back so other engines will see it. In
                # the future we may have additional codings, perhaps
                # ones that shouldn't auto propagate, so this sort of
                # test will become a bit more complicated.
                if o[0] == TaskIdOOB: self.kvs.put('.task', o)
                self.ciq.put((o, throttled))
                inFlight += 1
            else:
                logger.error('Unknown cylinder input tag: "%s" (%s)', tag, repr(o))
        self.kvs.close()

##################################################################### MAIN

if '__main__' == __name__:
    import argparse

    #sys.setcheckinterval(1000000)

    if len(sys.argv) > 1 and sys.argv[1] == '--engine':
        argp = argparse.ArgumentParser(description='Task execution engine.')
        argp.add_argument('--engine', action='store_true', help='Run in execution engine mode.')
        argp.add_argument('-n', '--node', type=str, help='Name of this engine node.')
        argp.add_argument('kvsserver', help='Address of kvs sever used to relay data to this execution engine.')
        args = argp.parse_args()
        # Stagger start randomly to throttle kvs connections
        time.sleep(random.random()*5.0)
        kvs = kvsstcp.KVSClient(args.kvsserver)
        context = kvs.view('.context')
        try:
            os.chdir(context.wd)
        except Exception as e:
            print('Failed to change working directory to "%s".'%context.wd, file=sys.stderr)
        context.setNode(args.node)
        logger = logging.getLogger('DisBatch Engine')
        lconf = {'format': '%(asctime)s %(levelname)-8s %(name)-15s: %(message)s', 'level': logging.INFO}
        lconf['filename'] = logfile(context, 'engine.log')
        logging.basicConfig(**lconf)
        logger.info('Starting engine %s (%d) on %s (%d) in %s.', context.node, context.nodeId, myHostname, myPid, os.getcwd())

        e = EngineBlock(kvs, context)

        def shutdown(s=None, f=None):
            logger.info('Engine shutting down.')
            for c in e.cylinders:
                if c.is_alive():
                    logger.info('forcing cylinder termination')
                    try:
                        c.terminate()
                    except OSError:
                        pass
            if s:
                logger.info('exiting on signal')
                sys.exit(1)
        signal.signal(signal.SIGTERM, shutdown)

        try:
            kvs.view('.shutdown')
            logger.info('got shutdown')
        except socket.error as r:
            logger.info('got socket error waiting on shutdown: %r'%r)
            pass
        finally:
            shutdown()
        kvs.close()
        logger.info('Remaining processes:\n' + SUB.check_output(['ps', 'fuhx']).decode('utf-8', 'ignore'))
        logger.info('exiting successfully')
        sys.exit(0)

    else:
        argp = argparse.ArgumentParser(description='Use batch resources to process a file of tasks, one task per line.')
        argp.add_argument('--fix-paths', action='store_true', help='Configure fixed path to script and modules.')
        argp.add_argument('-p', '--prefix', metavar='PATH', default=None, help='Prefix path and name for log and status files (default: ./TASKFILE_JOBID).')
        argp.add_argument('-l', '--logfile', metavar='FILE', default=None, type=argparse.FileType('w'), help='Log file.')
        argp.add_argument('--mailFreq', default=None, type=int, metavar='N', help='Send email every N task completions (default: 1). "--mailTo" must be given.')
        argp.add_argument('--mailTo', metavar='ADDR', default=None, help='Mail address for task completion notification(s).')
        argp.add_argument('-c', '--cpusPerTask', metavar='N', default=1, type=float, help='Number of cores used per task; may be fractional (default: 1).')
        argp.add_argument('-t', '--tasksPerNode', metavar='N', default=float('inf'), type=int, help='Maximum concurrently executing tasks per node (up to cores/cpusPerTask).')
        argp.add_argument('-E', '--env-resource', metavar='VAR', action='append', help=argparse.SUPPRESS) #'Assign comma-delimited resources specified in environment VAR across tasks (count should match -t)'
        argp.add_argument('-g', '--gpu', action='append_const', dest='env_resource', const='CUDA_VISIBLE_DEVICES,GPU_DEVICE_ORDINAL', help='Use assigned GPU resources')
        argp.add_argument('-k', '--retire-cmd', type=str, metavar='COMMAND', help="Shell command to run to retire a node (environment includes $NODE being retired, remaining $ACTIVE node list, $RETIRED node list; default based on batch system).")
        argp.add_argument('-K', '--no-retire', dest='retire_cmd', action='store_const', const='', help="Don't retire nodes from the batch system (e.g., if running as part of a larger job); equivalent to -k ''.")
        argp.add_argument('-s', '--ssh-node', type=str, action='append', metavar='HOST:COUNT', help="Run tasks over SSH on the given nodes (can be specified multiple times for additional hosts; equivalent to setting DISBATCH_SSH_NODELIST)")
        argp.add_argument('-r', '--resume-from', metavar='STATUSFILE', action='append', help='Read the status file from a previous run and skip any completed tasks (may be specified multiple times).')
        argp.add_argument('-R', '--retry', action='store_true', help='With -r, also retry any tasks which failed in previous runs (non-zero return).')
        argp.add_argument('--force-resume', action='store_true', help="With -r, proceed even if task commands/lines are different.")
        argp.add_argument('-e', '--exit-code', action='store_true', help='When any task fails, exit with non-zero status (default: only if disBatch itself fails)')
        argp.add_argument('-w', '--web', action='store_true', help='Enable web interface.')
        argp.add_argument('--kvsserver', nargs='?', default=True, metavar='HOST:PORT', help='Use a running KVS server.')
        source = argp.add_mutually_exclusive_group(required=True)
        source.add_argument('--taskcommand', default=None, metavar='COMMAND', help='Tasks will come from the command specified via the KVS server (passed in the environment).')
        source.add_argument('--taskserver', nargs='?', default=False, metavar='HOST:PORT', help='Tasks will come from the KVS server.')
        source.add_argument('taskfile', nargs='?', default=None, type=argparse.FileType('r', 1), help='File with tasks, one task per line ("-" for stdin)')
        args = argp.parse_args()

        if args.fix_paths:
            print('You must use --fix-paths without any other arguments.', file=sys.stderr)
            sys.exit(1)

        if args.mailFreq and not args.mailTo:
            argp.print_help()
            sys.exit(1)
        if not args.mailFreq and args.mailTo:
            args.mailFreq = 1

        if not args.kvsserver:
            args.kvsserver = args.taskserver
        elif args.taskserver is None: # --taskserver with no argument
            args.taskserver = args.kvsserver
        elif args.taskserver and args.kvsserver != args.taskserver:
            print('Cannot use different --kvsserver and --taskservers.', file=sys.stderr)
            sys.exit(1)

        # Try to find a batch context.
        if args.ssh_node:
            context = SSHContext(args.ssh_node)
        else:
            context = probeContext()
        if not context:
            print('Cannot determine batch execution environment.', file=sys.stderr)
            sys.exit(1)

        if args.prefix:
            context.name = args.prefix
        else:
            try:
                context.name = os.path.basename(args.taskfile.name).strip('<>')
            except AttributeError:
                context.name = 'disBatch'
            context.name += '_'+str(context.jobid)

        # Apply lesser of -c and -t limits
        context.cylinders = [ min(int(c / args.cpusPerTask), args.tasksPerNode) for c in context.cylinders ]
        # TODO: communicate to jobs how many CPUs they have available?

        context.envres = ','.join(args.env_resource).split(',') if args.env_resource else []

        if args.retire_cmd is not None:
            context.retireCmd = args.retire_cmd

        logger = logging.getLogger('DisBatch')
        lconf = {'format': '%(asctime)s %(levelname)-8s %(name)-15s: %(message)s', 'level': logging.INFO}
        if args.logfile:
            args.logfile.close()
            lconf['filename'] = args.logfile.name
        else:
            lconf['filename'] = logfile(context, 'driver.log')
        logging.basicConfig(**lconf)

        logger.info('Starting feeder (%d) on %s in %s.', myPid, myHostname, os.getcwd())
        logger.info('Context: %s', context)

        if args.kvsserver is True:
            # start our own
            kvsst = kvsstcp.KVSServerThread(socket.gethostname(), 0)
            kvsserver = '%s:%d'%kvsst.cinfo
            kvsinfotxt = logfile(context, 'kvsinfo.txt')
            with open(kvsinfotxt, 'w') as kvsi:
                kvsi.write(kvsserver)
            kvsenv = kvsst.env()
        else:
            # use one given (possibly via environment)
            kvsst = None
            kvsserver = args.kvsserver
            kvsenv = None

        logger.info('KVS Server: %s', kvsserver)
        kvs = kvsstcp.KVSClient(kvsserver)

        taskProcess = None
        if args.taskfile:
            taskSource = args.taskfile
        else:
            taskSource = KVSTaskSource(kvs)
            if args.taskcommand:
                taskProcess = TaskProcess(taskSource, args.taskcommand, shell=True, env=kvsenv, close_fds=True)

        tasks = taskGenerator(taskSource, context)

        if args.resume_from:
            tasks = statusTaskFilter(tasks, parseStatusFiles(*args.resume_from), args.retry, args.force_resume)

        if args.web:
            from kvsstcp import wskvsmu
            urlfile = logfile(context, 'url')
            wskvsmu.main(kvsserver, urlfile=open(urlfile, 'w'), monitorspec=':gpvw')

        context.launch(kvs)

        f = Driver(kvs, context, tasks, getattr(taskSource, 'resultkey', None), args.mailTo, args.mailFreq)
        try:
            while f.isAlive():
                context.poll()
                if not context.engines:
                    logger.warn('All engines terminated; shutting down')
                    break
                if taskProcess and taskProcess.r:
                    logger.warn('Task generator failed; forcing shutdown')
                    sys.exit(taskProcess.r)
                f.join(15)
        finally:
            try:
                logger.info("posting .shutdown")
                kvs.put('.shutdown', '', False)
            except:
                pass
            kvs.close()
            r = context.finish()
            if kvsst: kvsst.shutdown()
            if args.kvsserver is True:
                try:
                    os.unlink(kvsinfotxt)
                except OSError:
                    # may have been already removed -- can happen if multiple disbatch runs in same directory
                    pass
            if args.web: os.unlink(urlfile)

        if not r:
            print('Some engine processes failed -- please check the logs', file=sys.stderr)
            sys.exit(1)

        if args.exit_code and f.failed:
            print('Some tasks failed with non-zero exit codes -- please check the logs', file=sys.stderr)
            sys.exit(1)



# allsims testhod rsd
import numpy as np
import os, sys, time
import Corrfunc
# from Corrfunc.mocks.DDrppi_mocks import DDrppi_mocks
# from Corrfunc.utils import convert_3d_counts_to_cf, convert_rp_pi_counts_to_wp
# from Corrfunc.theory.DDrppi import 
from Corrfunc.theory import wp, xi, DDsmu, DDrppi
from scipy.special import legendre

def tpcf_multipole(s_mu_tcpf_result, mu_bins, order=0):
    r"""
    Calculate the multipoles of the two point correlation function
    after first computing `~halotools.mock_observables.s_mu_tpcf`.
    This is copied over from halotools. Original author was Duncan Campbell.
    Parameters
    ----------
    s_mu_tcpf_result : np.ndarray
        2-D array with the two point correlation function calculated in bins
        of :math:`s` and :math:`\mu`.  See `~halotools.mock_observables.s_mu_tpcf`.
    mu_bins : array_like
        array of :math:`\mu = \cos(\theta_{\rm LOS})`
        bins for which ``s_mu_tcpf_result`` has been calculated.
        Must be between [0,1].
    order : int, optional
        order of the multpole returned.
    Returns
    -------
    xi_l : np.array
        multipole of ``s_mu_tcpf_result`` of the indicated order.
    Examples
    --------
    For demonstration purposes we create a randomly distributed set of points within a
    periodic cube of length 250 Mpc/h.
    >>> Npts = 100
    >>> Lbox = 250.
    >>> x = np.random.uniform(0, Lbox, Npts)
    >>> y = np.random.uniform(0, Lbox, Npts)
    >>> z = np.random.uniform(0, Lbox, Npts)
    We transform our *x, y, z* points into the array shape used by the pair-counter by
    taking the transpose of the result of `numpy.vstack`. This boilerplate transformation
    is used throughout the `~halotools.mock_observables` sub-package:
    >>> sample1 = np.vstack((x,y,z)).T
    First, we calculate the correlation function using
    `~halotools.mock_observables.s_mu_tpcf`.
    >>> from halotools.mock_observables import s_mu_tpcf
    >>> s_bins  = np.linspace(0.01, 25, 10)
    >>> mu_bins = np.linspace(0, 1, 15)
    >>> xi_s_mu = s_mu_tpcf(sample1, s_bins, mu_bins, period=Lbox)
    Then, we can calculate the quadrapole of the correlation function:
    >>> xi_2 = tpcf_multipole(xi_s_mu, mu_bins, order=2)
    """

    # process inputs
    s_mu_tcpf_result = np.atleast_1d(s_mu_tcpf_result)
    mu_bins = np.atleast_1d(mu_bins)
    order = int(order)

    # calculate the center of each mu bin
    mu_bin_centers = (mu_bins[:-1]+mu_bins[1:])/(2.0)

    # get the Legendre polynomial of the desired order.
    Ln = legendre(order)

    # numerically integrate over mu
    result = (2.0*order + 1.0)/2.0 * np.sum(s_mu_tcpf_result * np.diff(mu_bins) *\
        (Ln(mu_bin_centers) + Ln(-1.0*mu_bin_centers)), axis=1)

    return result

def calc_xirppi_fast(x1, y1, z1, rpbins, pimax, 
    pi_bin_size, lbox, Nthread, num_cells = 20, x2 = None, y2 = None, z2 = None):  # all r assumed to be in h-1 mpc units. 
    start = time.time()
    if not isinstance(pimax, int):
        raise ValueError("pimax needs to be an integer")
    if not isinstance(pi_bin_size, int):
        raise ValueError("pi_bin_size needs to be an integer")
    if not pimax % pi_bin_size == 0:
        raise ValueError("pi_bin_size needs to be an integer divisor of pimax, current values are ", pi_bin_size, pimax)

    ND1 = float(len(x1))
    if x2 is not None:
        ND2 = len(x2)
        autocorr = 0
    else:
        autocorr = 1
        ND2 = ND1
    
    # single precision mode
    # to do: make this native 
    cf_start = time.time()
    rpbins = rpbins.astype(np.float32)
    pimax = np.float32(pimax)
    x1 = x1.astype(np.float32)
    y1 = y1.astype(np.float32)
    z1 = z1.astype(np.float32)
    lbox = np.float32(lbox)

    if autocorr == 1:    
        results = DDrppi(autocorr, Nthread, pimax, rpbins, x1, y1, z1,
            boxsize = lbox, periodic = True, max_cells_per_dim = num_cells)
        DD_counts = results['npairs']
    else:
        results = DDrppi(autocorr, Nthread, pimax, rpbins, x1, y1, z1, X2 = x2, Y2 = y2, Z2 = z2, 
            boxsize = lbox, periodic = True, max_cells_per_dim = num_cells)
        DD_counts = results['npairs']
    print("corrfunc took time ", time.time() - cf_start)

    DD_counts_new = np.array([np.sum(DD_counts[i:i+pi_bin_size]) for i in range(0, len(DD_counts), pi_bin_size)])
    DD_counts_new = DD_counts_new.reshape((len(rpbins) - 1, int(pimax/pi_bin_size)))

    # RR_counts_new = np.zeros((len(rpbins) - 1, int(pimax/pi_bin_size)))
    RR_counts_new = np.pi*(rpbins[1:]**2 - rpbins[:-1]**2)*pi_bin_size / lbox**3 * ND1 * ND2 * 2
    xirppi = DD_counts_new / RR_counts_new[:, None] - 1
    print("corrfunc took ", time.time() - start, "ngal ", len(x1))
    return xirppi


def calc_multipole_fast(x1, y1, z1, rpbins, 
    lbox, Nthread, num_cells = 20, x2 = None, y2 = None, z2 = None, orders = [0, 2, 4]):  # all r assumed to be in h-1 mpc units. 

    ND1 = float(len(x1))
    if x2 is not None:
        ND2 = len(x2)
        autocorr = 0
    else:
        autocorr = 1
        ND2 = ND1
    
    # single precision mode
    # to do: make this native 
    cf_start = time.time()
    rpbins = rpbins.astype(np.float32)
    x1 = x1.astype(np.float32)
    y1 = y1.astype(np.float32)
    z1 = z1.astype(np.float32)
    pos1 = np.array([x1, y1, z1]).T % lbox
    lbox = np.float32(lbox)

    # mu_bins = np.linspace(0, 1, 20)

    # if autocorr == 1: 
    #     xi_s_mu = s_mu_tpcf(pos1, rpbins, mu_bins, period = lbox, num_threads = Nthread)
    #     print("halotools ", xi_s_mu)
    # else:
    #     xi_s_mu = s_mu_tpcf(pos1, rpbins, mu_bins, sample2 = np.array([x2, y2, z2]).T % lbox, period = lbox, num_threads = Nthread)

    nbins_mu = 20
    if autocorr == 1: 
        results = DDsmu(autocorr, Nthread, rpbins, 1, nbins_mu, x1, y1, z1, periodic = True, boxsize = lbox, max_cells_per_dim = num_cells)
        DD_counts = results['npairs']
    else:
        results = DDsmu(autocorr, Nthread, rpbins, 1, nbins_mu, x1, y1, z1, X2 = x2, Y2 = y2, Z2 = z2, 
            periodic = True, boxsize = lbox, max_cells_per_dim = num_cells)
        DD_counts = results['npairs']
    DD_counts = DD_counts.reshape((len(rpbins) - 1, nbins_mu))

    mu_bins = np.linspace(0, 1, nbins_mu+1)
    RR_counts = 2*np.pi/3*(rpbins[1:, None]**3 - rpbins[:-1, None]**3)*(mu_bins[None, 1:] - mu_bins[None, :-1]) / lbox**3 * ND1 * ND2 * 2

    xi_s_mu = DD_counts / RR_counts - 1

    xi_array = []
    for neworder in orders:
        # print(neworder, rpbins, tpcf_multipole(xi_s_mu, mu_bins, order = neworder))
        xi_array += [tpcf_multipole(xi_s_mu, mu_bins, order=neworder)]
    xi_array = np.concatenate(xi_array)

    return xi_array


def calc_wp_fast(x1, y1, z1, rpbins, pimax, 
    lbox, Nthread, num_cells = 30, x2 = None, y2 = None, z2 = None):  # all r assumed to be in h-1 mpc units. 
    if not isinstance(pimax, int):
        raise ValueError("pimax needs to be an integer")

    ND1 = float(len(x1))
    if x2 is not None:
        ND2 = len(x2)
        autocorr = 0
    else:
        autocorr = 1
        ND2 = ND1

    # single precision mode
    # to do: make this native 
    cf_start = time.time()
    rpbins = rpbins.astype(np.float32)
    pimax = np.float32(pimax)
    x1 = x1.astype(np.float32)
    y1 = y1.astype(np.float32)
    z1 = z1.astype(np.float32)
    lbox = np.float32(lbox)

    if autocorr == 1:    
        print("sample size", len(x1))
        results = DDrppi(autocorr, Nthread, pimax, rpbins, x1, y1, z1,
            boxsize = lbox, periodic = True, max_cells_per_dim = num_cells)
        DD_counts = results['npairs']
    else:
        print("sample size", len(x1), len(x2))
        x2 = x2.astype(np.float32)
        y2 = y2.astype(np.float32)
        z2 = z2.astype(np.float32)
        results = DDrppi(autocorr, Nthread, pimax, rpbins, x1, y1, z1, X2 = x2, Y2 = y2, Z2 = z2, 
            boxsize = lbox, periodic = True, max_cells_per_dim = num_cells)
        DD_counts = results['npairs']
    print("corrfunc took time ", time.time() - cf_start)
    DD_counts = DD_counts.reshape((len(rpbins) - 1, int(pimax)))

    # RR_counts = np.zeros((len(rpbins) - 1, int(pimax)))
    # for i in range(len(rpbins) - 1):
    RR_counts = np.pi*(rpbins[1:]**2 - rpbins[:-1]**2) / lbox**3 * ND1 * ND2 * 2
    xirppi = DD_counts / RR_counts[:, None] - 1

    return 2*np.sum(xirppi, axis = 1)





import os
import socket

# The line protocol is very simple:
#
#  4 byte operation ('clos', 'dump', 'get_', 'mkey', 'put_', 'view')
#
# 'clos': No additional argument
#
# 'dump': No additional argument
#
# 'get_': One key argument, expects a value argument in reply.
#
# 'mkey' (monitor key): Two key arguments. The second may have a ': ...'
#     suffix indicating events to be monitored.
#
# 'put_': One key argument followed by one value argument.
# 
# 'view': One key argument, expects a value argument in reply.
#
# Key representation:
#     10 bytes: A 10 character string (ascii, not null terminated) with the base 10
#               representation of the byte length of the key string
#     length bytes: the key string
#
# Value representatin:
#     4 bytes: coding scheme.
#     10 bytes: A 10 character string (ascii, not null terminated) with the base 10
#               representation of the byte length of the argument
#     length bytes: the string representing the key
#
# Notes:
#
# 1) Coding schemes for values is a work in progress.
#

AsciiLenChars = 10 

def AsciiLenFormat(n):
    assert(n <= 9999999999)
    return str(n).encode('ascii').rjust(AsciiLenChars)

if hasattr(socket, "MSG_WAITALL") and os.uname()[0] != 'Darwin':
    # MSG_WAITALL on OSX ends up blocking if the tcp buffer is not big enough for the entire message: don't use it
    def recvall(s, n):
        if not n: return b''
        r = s.recv(n, socket.MSG_WAITALL)
        if len(r) < n: raise socket.error('Connection dropped')
        return r
else:
    def recvall(s, n):
        '''Wrapper to deal with partial recvs when we know there are N bytes to be had.'''
        d = b''
        while n:
            b = s.recv(n)
            if not b: raise socket.error('Connection dropped')
            d += b
            n -= len(b)
        return d



#!/usr/bin/env python
from distutils.core import setup
setup(name='kvsstcp', version='0',
    description="Key value storage server",
    long_description=open("Readme.md").read(),
    packages=['kvsstcp'],
    package_dir={'kvsstcp':'.'},
    package_data={'kvsstcp':['wskvspage.html']})



#! /usr/bin/env python

import os
import time
import sys

import numpy as np
import argparse
import emcee
import yaml

from likelihood import PowerData
from abacusnbody.hod.abacus_hod import AbacusHOD

DEFAULTS = {}
DEFAULTS['path2config'] = 'config/abacus_hod.yaml'

class SampleFileUtil(object):
    """
    Util for handling sample files.
    Copied from Andrina's code.

    :param filePrefix: the prefix to use
    :param reuseBurnin: True if the burn in data from a previous run should be used
    """

    def __init__(self, filePrefix, carry_on=False):
        self.filePrefix = filePrefix
        if carry_on:
            mode = 'a'
        else:
            mode = 'w'
        self.samplesFile = open(self.filePrefix + '.txt', mode)
        self.probFile = open(self.filePrefix + 'prob.txt', mode)

    def persistSamplingValues(self, pos, prob):
        self.persistValues(self.samplesFile, self.probFile, pos, prob)

    def persistValues(self, posFile, probFile, pos, prob):
        """
        Writes the walker positions and the likelihood to the disk
        """
        posFile.write("\n".join(["\t".join([str(q) for q in p]) for p in pos]))
        posFile.write("\n")
        posFile.flush()

        probFile.write("\n".join([str(p) for p in prob]))
        probFile.write("\n")
        probFile.flush();

    def close(self):
        self.samplesFile.close()
        self.probFile.close()

    def __str__(self, *args, **kwargs):
        return "SampleFileUtil"

class DumPool(object):
    def __init__(self):
        pass

    def is_master(self):
        return True

    def close(self):
        pass
    
def time_lnprob(params, param_mapping, param_tracer, Data, Ball):
    print('   ==========================================')
    print("   | Calculating likelihood evaluation time |")
    print('   ==========================================')

    # run once without timing since numba needs to compile
    lnprob(params[:, 0], params, param_mapping, param_tracer, Data, Ball)
    
    timing = np.zeros(10)
    for i in range(10):
        print('Test ',i,' of 9')
        start = time.time()
        if i<5:
            lnprob(params[:, 0]+i*0.1*params[:, 3], params, param_mapping, param_tracer, Data, Ball)
        else:
            lnprob(params[:, 0]-(i-4)*0.1*params[:, 3], params, param_mapping, param_tracer, Data, Ball)
        finish = time.time()
        timing[i] = finish-start
                
    mean = np.mean(timing)
    print('============================================================================')
    print('mean computation time: ', mean)
    stdev = np.std(timing)
    print('standard deviation : ', stdev)
    print('============================================================================')
    return

def inrange(p, params):
    return np.all((p<=params[:, 2]) & (p>=params[:, 1]))

def lnprob(p, params, param_mapping, param_tracer, Data, Ball):
    if inrange(p, params):
        # read the parameters 
        for key in param_mapping.keys():
            mapping_idx = param_mapping[key]
            tracer_type = param_tracer[key]
            #tracer_type = param_tracer[params[mapping_idx, -1]]
            Ball.tracers[tracer_type][key] = p[mapping_idx]
            print(key, Ball.tracers[tracer_type][key])
            
        # pass them to the mock dictionary
        mock_dict = Ball.run_hod(Ball.tracers, Ball.want_rsd, Nthread = 64)
        clustering = Ball.compute_xirppi(mock_dict, Ball.rpbins, Ball.pimax, Ball.pi_bin_size, Nthread = 16)
        lnP = Data.compute_likelihood(clustering)
    else:
        lnP = -np.inf
    return lnP


def main(path2config, time_likelihood):

    # load the yaml parameters
    config = yaml.load(open(path2config))
    sim_params = config['sim_params']
    HOD_params = config['HOD_params']
    clustering_params = config['clustering_params']
    data_params = config['data_params']
    ch_config_params = config['ch_config_params']
    fit_params = config['fit_params']    
    
    # create a new abacushod object and load the subsamples
    newBall = AbacusHOD(sim_params, HOD_params, clustering_params)

    # read data parameters
    newData = PowerData(data_params, HOD_params)

    # parameters to fit
    nparams = len(fit_params.keys())
    param_mapping = {}
    param_tracer = {}
    params = np.zeros((nparams, 4))
    for key in fit_params.keys():
        mapping_idx = fit_params[key][0]
        tracer_type = fit_params[key][-1]
        param_mapping[key] = mapping_idx
        param_tracer[key] = tracer_type
        params[mapping_idx, :] = fit_params[key][1:-1]
        
    # Make path to output
    if not os.path.isdir(os.path.expanduser(ch_config_params['path2output'])):
        try:
            os.makedirs(os.path.expanduser(ch_config_params['path2output']))
        except:
            pass

    # MPI option
    if ch_config_params['use_mpi']:
        from schwimmbad import MPIPool
        pool = MPIPool()
        print("Using MPI")
        pool_use = pool
    else:
        pool = DumPool()
        print("Not using MPI")
        pool_use = None
        
    if not pool.is_master():
        pool.wait()
        sys.exit(0)
    
    # just time the likelihood calculation
    if time_likelihood:
        time_lnprob(params, param_mapping, param_tracer, newData, newBall)
        return
    
    # emcee parameters
    nwalkers = nparams * ch_config_params['walkersRatio']
    nsteps = ch_config_params['burninIterations'] + ch_config_params['sampleIterations']

    # where to record
    prefix_chain = os.path.join(os.path.expanduser(ch_config_params['path2output']),
                                ch_config_params['chainsPrefix'])

    # fix initial conditions
    found_file = os.path.isfile(prefix_chain+'.txt')
    if (not found_file) or (not ch_config_params['rerun']):
        p_initial = params[:, 0] + np.random.normal(size=(nwalkers, nparams)) * params[:, 3][None, :]
        nsteps_use = nsteps
    else:
        print("Restarting from a previous run")
        old_chain = np.loadtxt(prefix_chain+'.txt')
        p_initial = old_chain[-nwalkers:,:]
        nsteps_use = max(nsteps-len(old_chain) // nwalkers, 0)

    # initializing sampler
    chain_file = SampleFileUtil(prefix_chain, carry_on=ch_config_params['rerun'])
    sampler = emcee.EnsembleSampler(nwalkers, nparams, lnprob, args=(params, param_mapping, param_tracer, newData, newBall), pool=pool_use)
    start = time.time()
    print("Running %d samples" % nsteps_use)

    # record every iteration
    counter = 1
    for pos, prob, _ in sampler.sample(p_initial, iterations=nsteps_use):
        if pool.is_master():
            print('Iteration done. Persisting.')
            chain_file.persistSamplingValues(pos, prob)

            if counter % 10:
                print(f"Finished sample {counter}")
        counter += 1

    pool.close()
    end = time.time()
    print("Took ",(end - start)," seconds")


class ArgParseFormatter(argparse.RawDescriptionHelpFormatter, argparse.ArgumentDefaultsHelpFormatter):
    pass


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=ArgParseFormatter)
    parser.add_argument('--path2config', dest='path2config', type=str, help='Path to config file.', default=DEFAULTS['path2config'])
    parser.add_argument('--time_likelihood', dest='time_likelihood',  help='Times the likelihood calculations', action='store_true')
    
    args = vars(parser.parse_args())    
    main(**args)




'''
This module contains the ASDF extensions that allow the asdf Python package
to read Abacus ASDF files that use Blosc compression internally.

There are two classes here: an Extension subclass, and a Compressor subclass.
The Extension is registered with ASDF via a setuptools "entry point" in setup.py.
It contains the reference to the Compressor subclass that knows how to
handle Blosc compression.
'''

import time
import struct

import numpy as np
import blosc
from asdf.extension import Extension, Compressor

import asdf
def _monkey_patch(*args,**kwargs):
    raise Exception("Please use abacusnbody.data.asdf.set_nthreads(nthreads)")
    
asdf.compression.set_decompression_options = _monkey_patch

def set_nthreads(nthreads):
    blosc.set_nthreads(nthreads)
    

class BloscCompressor(Compressor):
    '''
    An implementation of Blosc compression, as used by Abacus.
    '''
        
    @property
    def label(self):
        '''
        The string labels in the binary block headers
        that indicate Blosc compression
        '''
        return b'blsc'
    
    
    def compress(self, data, **kwargs):
        '''Useful compression kwargs:
        nthreads
        compression_block_size
        blosc_block_size
        shuffle
        typesize
        cname
        clevel
        '''
        # Blosc code probably assumes contiguous buffer
        assert data.contiguous
        
        nthreads = kwargs.pop('nthreads', 1)
        compression_block_size = kwargs.pop('compression_block_size',1<<22)
        blosc_block_size = kwargs.pop('blosc_block_size', 512*1024)
        typesize = kwargs.pop('typesize','auto')  # dtype size in bytes, e.g. 8 for int64
        clevel = kwargs.pop('clevel',1)  # compression level, usually only need lowest for zstd
        cname = kwargs.pop('cname','zstd')  # compressor name, default zstd, good performance/compression tradeoff
        
        shuffle = kwargs.pop('shuffle', None)
        if shuffle == 'shuffle':
            shuffle = blosc.SHUFFLE
        elif shuffle == 'bitshuffle':
            shuffle = blosc.BITSHUFFLE
        elif shuffle == None:
            shuffle = blosc.NOSHUFFLE
        else:
            raise ValueError(shuffle)
        
        blosc.set_nthreads(nthreads)
        blosc.set_blocksize(blosc_block_size)
        
        if typesize == 'auto':
            this_typesize = data.itemsize
        else:
            this_typesize = typesize
        #assert this_typesize != 1
        
        nelem = compression_block_size // data.itemsize
        for i in range(0,len(data),nelem):
            compressed = blosc.compress(data[i:i+nelem], typesize=this_typesize, clevel=clevel, shuffle=shuffle, cname=cname,
                                    **kwargs)
            header = struct.pack('!I', len(compressed))
            # TODO: this probably triggers a data copy, feels inefficient. Probably have to add output array arg to blosc to fix
            yield header + compressed
    
    
    def decompress(self, blocks, out, **kwargs):
        '''Useful decompression kwargs:
        nthreads
        '''
        # TODO: controlled globally for now
        #nthreads = kwargs.pop('nthreads',1)
        #blosc.set_nthreads(nthreads)
        
        _size = 0
        _pos = 0
        _buffer = None
        _partial_len = b''
        
        decompression_time = 0.
        bytesout = 0
        
        # Blosc code probably assumes contiguous buffer
        if not out.contiguous:
            raise ValueError(out.contiguous)
        
        # get the out address
        out = np.frombuffer(out, dtype=np.uint8).ctypes.data
        
        for block in blocks:
            block = memoryview(block).cast('c')
            try:
                block = block.toreadonly()  # python>=3.8 only
            except AttributeError:
                pass
            
            if not block.contiguous:
                raise ValueError(block.contiguous)
                
            while len(block):
                if not _size:
                    # Don't know the (compressed) length of this block yet
                    if len(_partial_len) + len(block) < 4:
                        _partial_len += block
                        break  # we've exhausted the data
                    if _partial_len:
                        # If we started to fill a len key, finish filling it
                        remaining = 4-len(_partial_len)
                        if remaining:
                            _partial_len += block[:remaining]
                            block = block[remaining:]
                        _size = struct.unpack('!I', _partial_len)[0]
                        _partial_len = b''
                    else:
                        # Otherwise just read the len key directly
                        _size = struct.unpack('!I', block[:4])[0]
                        block = block[4:]

                if len(block) < _size or _buffer is not None:
                    # If we have a partial block, or we're already filling a buffer, use the buffer
                    if _buffer is None:
                        _buffer = np.empty(_size, dtype=np.byte)  # use numpy instead of bytearray so we can avoid zero initialization
                        _pos = 0
                    newbytes = min(_size - _pos, len(block))  # don't fill past the buffer len!
                    _buffer[_pos:_pos+newbytes] = np.frombuffer(block[:newbytes], dtype=np.byte)
                    _pos += newbytes
                    block = block[newbytes:]

                    if _pos == _size:
                        start = time.perf_counter()
                        n_thisout = blosc.decompress_ptr(memoryview(_buffer), out + bytesout, **kwargs)
                        decompression_time += time.perf_counter() - start
                        bytesout += n_thisout
                        _buffer = None
                        _size = 0
                else:
                    # We have at least one full block
                    start = time.perf_counter()
                    n_thisout = blosc.decompress_ptr(memoryview(block[:_size]), out + bytesout, **kwargs)
                    decompression_time += time.perf_counter() - start
                    bytesout += n_thisout
                    block = block[_size:]
                    _size = 0

        return bytesout
       

class AbacusExtension(Extension):
    '''
    An ASDF Extension that deals with Abacus types and formats.
    Currently only implements Blosc compression.
    '''
    
    @property
    def extension_uri(self):
        """
        Get the URI of the extension to the ASDF Standard implemented
        by this class.  Note that this may not uniquely identify the
        class itself.

        Returns
        -------
        str
        """
        return "asdf://abacusnbody.org/extensions/abacus-0.0.1"
    
    @property
    def compressors(self):
        '''
        Return the Compressors implemented in this extension
        '''
        return [BloscCompressor()]



#! /usr/bin/env python

import os
import time
import sys

import numpy as np
import argparse
from dynesty import NestedSampler
import yaml
import dill
from scipy import stats

from likelihood import PowerData
from abacusnbody.hod.abacus_hod import AbacusHOD

DEFAULTS = {}
DEFAULTS['path2config'] = 'config/boss_lrg_wp.yaml'


def lnprob(p, param_mapping, param_tracer, Data, Ball):
    # read the parameters 
    for key in param_mapping.keys():
        mapping_idx = param_mapping[key]
        tracer_type = param_tracer[key]
        #tracer_type = param_tracer[params[mapping_idx, -1]]
        Ball.tracers[tracer_type][key] = p[mapping_idx]
        
    # pass them to the mock dictionary
    mock_dict = Ball.run_hod(Ball.tracers, Ball.want_rsd, Nthread = 64)

    clustering = Ball.compute_wp(mock_dict, Ball.rpbins, Ball.pimax, Ball.pi_bin_size, Nthread = 16)

    lnP = Data.compute_likelihood(theory_density)

    return lnP

# prior transform function
def prior_transform(u, params_hod, params_hod_initial_range):
    return stats.norm.ppf(u, loc = params_hod, scale = params_hod_initial_range)

def main(path2config):

    # load the yaml parameters
    config = yaml.load(open(path2config))
    sim_params = config['sim_params']
    HOD_params = config['HOD_params']
    clustering_params = config['clustering_params']
    data_params = config['data_params']
    dynesty_config_params = config['dynesty_config_params']
    fit_params = config['dynesty_fit_params']    
    
    # create a new abacushod object and load the subsamples
    newBall = AbacusHOD(sim_params, HOD_params, clustering_params)

    # read data parameters
    newData = wp_Data(data_params, HOD_params)

    # parameters to fit
    nparams = len(fit_params.keys())
    param_mapping = {}
    param_tracer = {}
    params = np.zeros((nparams, 2))
    for key in fit_params.keys():
        mapping_idx = fit_params[key][0]
        tracer_type = fit_params[key][-1]
        param_mapping[key] = mapping_idx
        param_tracer[key] = tracer_type
        params[mapping_idx, :] = fit_params[key][1:-1]

    # Make path to output
    if not os.path.isdir(os.path.expanduser(dynesty_config_params['path2output'])):
        try:
            os.makedirs(os.path.expanduser(dynesty_config_params['path2output']))
        except:
            pass
        
    # dynesty parameters
    nlive = dynesty_config_params['nlive']
    maxcall = dynesty_config_params['maxcall']
    method = dynesty_config_params['method']
    bound = dynesty_config_params['bound']

    # where to record
    prefix_chain = os.path.join(os.path.expanduser(dynesty_config_params['path2output']),
                                dynesty_config_params['chainsPrefix'])

    # initiate sampler
    found_file = os.path.isfile(prefix_chain+'.dill')
    if (not found_file) or (not dynesty_config_params['rerun']):

        # initialize our nested sampler
        sampler = NestedSampler(lnprob, prior_transform, nparams, 
            logl_args = [param_mapping, param_tracer, newData, newBall], 
            ptform_args = [params[:, 0], params[:, 1]], 
            nlive=nlive, sample = method, rstate = np.random.RandomState(dynesty_config_params['rseed']))
            # first_update = {'min_eff': 20})

    else:
        # load sampler to continue the run
        with open(prefix_chain+'.dill', "rb") as f:
            sampler = dill.load(f)
        sampler.rstate = np.load(prefix_chain+'_results.npz')['rstate']
    print("run sampler")

    sampler.run_nested(maxcall = maxcall)

    # save sampler itself
    with open(prefix_chain+'.dill', "wb") as f:
         dill.dump(sampler, f)
    res1 = sampler.results
    np.savez(prefix_chain+'_results.npz', res = res1, rstate = np.random.get_state())

class ArgParseFormatter(argparse.RawDescriptionHelpFormatter, argparse.ArgumentDefaultsHelpFormatter):
    pass


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=ArgParseFormatter)
    parser.add_argument('--path2config', dest='path2config', type=str, help='Path to config file.', default=DEFAULTS['path2config'])
    
    args = vars(parser.parse_args())    
    main(**args)




import tempfile
import filecmp
from pathlib import Path
import numbers

import pytest
from astropy.table import Table
import numpy as np

curdir = Path(__file__).parent
refdir = curdir / 'ref'
EXAMPLE_SIM = curdir / 'Mini_N64_L32'
HALOS_OUTPUT_UNCLEAN = refdir / 'test_halos_unclean.asdf'
PARTICLES_OUTPUT_UNCLEAN = refdir / 'test_subsamples_unclean.asdf'
HALOS_OUTPUT_CLEAN = refdir / 'test_halos_clean.asdf'
PARTICLES_OUTPUT_CLEAN = refdir / 'test_subsamples_clean.asdf'
PACK9_OUTPUT = refdir / 'test_pack9.asdf'
PACK9_PID_OUTPUT = refdir / 'test_pack9_pid.asdf'

def test_halos_unclean(tmp_path):
    '''Test loading a base (uncleaned) halo catalog
    '''

    from abacusnbody.data.compaso_halo_catalog import CompaSOHaloCatalog

    cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000', subsamples=True, fields='all', cleaned=False)

    # to regenerate reference
    #ref = cat.halos
    #ref.write(HALOS_OUTPUT_UNCLEAN, all_array_storage='internal', all_array_compression='blsc')

    ref = Table.read(HALOS_OUTPUT_UNCLEAN)

    halos = cat.halos
    for col in ref.colnames:
        if issubclass(ref[col].dtype.type, numbers.Integral):
            assert np.all(halos[col] == ref[col])
        else:
            assert np.allclose(halos[col], ref[col])

    assert halos.meta == ref.meta

def test_halos_clean(tmp_path):
    '''Test loading a base (uncleaned) halo catalog
    '''

    from abacusnbody.data.compaso_halo_catalog import CompaSOHaloCatalog

    cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000', subsamples=True, fields='all', cleaned=True)

    # to regenerate reference
    #ref = cat.halos
    #ref.write(HALOS_OUTPUT_CLEAN, all_array_storage='internal', all_array_compression='blsc')

    ref = Table.read(HALOS_OUTPUT_CLEAN)

    halos = cat.halos
    for col in ref.colnames:
        if issubclass(ref[col].dtype.type, numbers.Integral):
            assert np.all(halos[col] == ref[col])
        else:
            assert np.allclose(halos[col], ref[col])

    # all haloindex values should point to this slab
    assert np.all((halos['haloindex']/1e12).astype(int) == cat.header['FullStepNumber'])
    # ensure that all deleted halos in ref are marked as merged in EXAMPLE_SIM
    assert np.all(halos['is_merged_to'][ref['N']==0] != -1)
    # no deleted halos in ref should have merged particles in EXAMPLE_SIM
    assert np.all(halos['N_merge'][ref['N']==0] == 0)

    assert halos.meta == ref.meta

def test_subsamples_unclean(tmp_path):
    '''Test loading particle subsamples
    '''

    from abacusnbody.data.compaso_halo_catalog import CompaSOHaloCatalog
    
    cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000', subsamples=dict(A=True), fields='all', cleaned=False)
    lenA = len(cat.subsamples)
    assert lenA == 2975
    assert cat.subsamples.colnames == ['pos', 'vel']
    
    cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000', subsamples=dict(B=True), fields='all', cleaned=False)
    lenB = len(cat.subsamples)
    assert lenB == 7082

    cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000', subsamples=True, fields='all', cleaned=False)
    
    assert len(cat.subsamples) == lenA + lenB

    # to regenerate reference
    #ref = cat.subsamples
    #import asdf; asdf.compression.set_compression_options(typesize='auto')
    #ref.write(PARTICLES_OUTPUT_UNCLEAN, format='asdf', all_array_storage='internal', all_array_compression='blsc')

    ref = Table.read(PARTICLES_OUTPUT_UNCLEAN)

    ss = cat.subsamples
    for col in ref.colnames:
        if issubclass(ref[col].dtype.type, numbers.Integral):
            assert np.all(ss[col] == ref[col])
        else:
            assert np.allclose(ss[col], ref[col])

    assert cat.subsamples.meta == ref.meta

def test_subsamples_clean(tmp_path):
    '''Test loading particle subsamples
    '''

    from abacusnbody.data.compaso_halo_catalog import CompaSOHaloCatalog

    cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000', subsamples=True, fields='all', cleaned=True)

    # to regenerate reference
    #ref = cat.subsamples
    #import asdf; asdf.compression.set_compression_options(typesize='auto')
    #ref.write(PARTICLES_OUTPUT_CLEAN, format='asdf', all_array_storage='internal', all_array_compression='blsc')

    ref = Table.read(PARTICLES_OUTPUT_CLEAN)

    ss = cat.subsamples
    for col in ref.colnames:
        if issubclass(ref[col].dtype.type, numbers.Integral):
            assert np.all(ss[col] == ref[col])
        else:
            assert np.allclose(ss[col], ref[col])

    # total number of particles in ref should be equal to the sum total of npout{AB} in EXAMPLE_SIM
    assert len(ref) == np.sum(cat.halos['npoutA']) + np.sum(cat.halos['npoutB'])

    assert cat.subsamples.meta == ref.meta

def test_field_subset_loading():
    '''Test loading a subset of halo catalog columns
    '''
    from abacusnbody.data.compaso_halo_catalog import CompaSOHaloCatalog

    cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000', fields=['N','x_com'])
    assert set(cat.halos.colnames) == set(['N','x_com'])


def test_one_halo_info():
    '''Test loading a single halo_info file
    '''
    from abacusnbody.data.compaso_halo_catalog import CompaSOHaloCatalog

    cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000'/'halo_info'/'halo_info_000.asdf',
        subsamples=True)
    assert len(cat.halos) == 127
    assert len(cat.subsamples) == 3209 #9306


def test_halo_info_list():
    '''Test list of halo infos
    '''
    from abacusnbody.data.compaso_halo_catalog import CompaSOHaloCatalog

    cat = CompaSOHaloCatalog(
        [EXAMPLE_SIM/'halos'/'z0.000'/'halo_info'/'halo_info_000.asdf',
         EXAMPLE_SIM/'halos'/'z0.000'/'halo_info'/'halo_info_001.asdf'],
        subsamples=True)
    assert len(cat.halos) == 281
    assert len(cat.subsamples) == 6900 #19555

    # check fail on dups
    with pytest.raises(ValueError):
        cat = CompaSOHaloCatalog(
        [EXAMPLE_SIM/'halos'/'z0.000'/'halo_info'/'halo_info_000.asdf',
         EXAMPLE_SIM/'halos'/'z0.000'/'halo_info'/'halo_info_000.asdf'])


def test_unpack_bits():
    '''Test unpack_bits
    '''

    from abacusnbody.data.compaso_halo_catalog import CompaSOHaloCatalog
    from abacusnbody.data.bitpacked import PID_FIELDS

    cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000', subsamples=True, unpack_bits=True, fields='N')
    assert set(PID_FIELDS) <= set(cat.subsamples.colnames)  # check subset

    cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000', subsamples=True, unpack_bits='density', fields='N')
    assert 'density' in cat.subsamples.colnames
    assert 'lagr_pos' not in cat.subsamples.colnames  # too many?

    # bad bits field name
    with pytest.raises(ValueError):
        cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000', subsamples=True, unpack_bits=['blah'], fields='N')


def test_filter_func():
    '''Test CHC filter_func
    '''
    
    from abacusnbody.data.compaso_halo_catalog import CompaSOHaloCatalog

    cat = CompaSOHaloCatalog(EXAMPLE_SIM/'halos'/'z0.000', fields=['N','x_L2com'],
                            filter_func = lambda c: c['N'] > 100,
                            subsamples=True)
    assert (cat.halos['N'] > 100).all()
    assert len(cat.halos) == 146
    assert len(cat.subsamples) == 7193


def test_pack9():
    '''Test reading a pack9 timeslice file
    '''
    from abacusnbody.data.read_abacus import read_asdf
    fn = EXAMPLE_SIM/'slices'/'z0.000'/'L0_pack9'/'slab000.L0.pack9.asdf'
    p = read_asdf(fn, load=('pos','vel'),
                dtype=np.float32)
    
    #p.write(PACK9_OUTPUT, format='asdf', all_array_compression='blsc')
    ref = Table.read(PACK9_OUTPUT)
    
    for k in ref.colnames:
        assert np.all(p[k] == ref[k])
    assert p.meta == ref.meta
    
    p = read_asdf(fn, dtype=np.float32)
    assert sorted(p.colnames) == ['pos','vel']

    # pid checks
    pidfn = EXAMPLE_SIM/'slices'/'z0.000'/'L0_pack9_pid'/'slab000.L0.pack9.pid.asdf'
    p = read_asdf(pidfn, load=('aux','pid','lagr_pos','tagged','density','lagr_idx'))
    
    #p.write(PACK9_PID_OUTPUT, format='asdf', all_array_compression='blsc')
    ref = Table.read(PACK9_PID_OUTPUT)
    
    for k in ref.colnames:
        assert np.all(p[k] == ref[k])
    assert p.meta == ref.meta
    
    p = read_asdf(pidfn, dtype=np.float32)
    assert p.colnames == ['pid']

    
def test_halo_lc():
    '''Test loading halo light cones
    '''
    
    from abacusnbody.data.compaso_halo_catalog import CompaSOHaloCatalog

    cat = CompaSOHaloCatalog(curdir / 'halo_light_cones/AbacusSummit_base_c000_ph001-abridged/z2.250/',
                             fields='all',
                             subsamples=True)
    assert(cat.halo_lc == True)
    
    HALO_LC_CAT = refdir / 'halo_lc_cat.asdf'
    HALO_LC_SUBSAMPLES = refdir / 'halo_lc_subsample.asdf'
    
    # generate reference
    #ref = cat.halos
    #ref.write(HALO_LC_CAT, format='asdf', all_array_storage='internal', all_array_compression='blsc')
    
    #ref = cat.subsamples
    #ref.write(HALO_LC_SUBSAMPLES, format='asdf', all_array_storage='internal', all_array_compression='blsc')
    
    ref = Table.read(HALO_LC_CAT)
    halos = cat.halos
    for col in ref.colnames:
        if issubclass(ref[col].dtype.type, numbers.Integral):
            assert np.all(halos[col] == ref[col])
        else:
            assert np.allclose(halos[col], ref[col])        
    assert halos.meta == ref.meta
    
    ref = Table.read(HALO_LC_SUBSAMPLES)
    ss = cat.subsamples
    for col in ref.colnames:
        if issubclass(ref[col].dtype.type, numbers.Integral):
            assert np.all(ss[col] == ref[col])
        else:
            assert np.allclose(ss[col], ref[col])
            
    assert ss.meta == ref.meta
